from __future__ import annotations

from collections.abc import Iterator
from typing import Generic, TypeVar

from cle.utils import key_bisect_find, key_bisect_insort_left

from .region import Region

R = TypeVar("R", bound=Region)


class Regions(Generic[R]):
    """
    A container class acting as a list of regions (sections or segments). Additionally, it keeps an sorted list of
    all regions that are mapped into memory to allow fast lookups.

    We assume none of the regions overlap with others.
    """

    def __init__(self, lst: list[R] | None = None):
        self._list = lst if lst is not None else []

        if self._list:
            self._sorted_list: list[R] = self._make_sorted(self._list)
        else:
            self._sorted_list = []

    @property
    def raw_list(self) -> list[R]:
        """
        Get the internal list. Any change to it is not tracked, and therefore _sorted_list will not be updated.
        Therefore you probably does not want to modify the list.

        :return:  The internal list container.
        :rtype:   list
        """

        return self._list

    @property
    def max_addr(self) -> int | None:
        """
        Get the highest address of all regions.

        :return: The highest address of all regions, or None if there is no region available.
        :rtype:   int or None
        """

        if self._sorted_list:
            return self._sorted_list[-1].max_addr
        return None

    def __getitem__(self, idx: int) -> R:
        return self._list[idx]

    def __setitem__(self, idx: int, item: R) -> None:
        self._list[idx] = item

        # update self._sorted_list
        self._sorted_list = self._make_sorted(self._list)

    def __iter__(self) -> Iterator[R]:
        return iter(self._list)

    def __len__(self) -> int:
        return len(self._list)

    def __repr__(self):
        return f"<Regions: {repr(self._list)}>"

    def _rebase(self, delta):
        """
        Does regions rebasing to other base address.
        Modifies state of each internal object, so the list reference doesn't need to be updated,
        the same is also valid for sorted list as operation preserves the ordering.

        :param delta: Delta offset between an old and a new image bases
        :type delta: int
        """
        for x in self._list:
            x._rebase(delta)

    def append(self, region: R):
        """
        Append a new Region instance into the list.

        :param region: The region to append.
        """
        self._list.append(region)

        if self._is_region_mapped(region):
            key_bisect_insort_left(self._sorted_list, region, keyfunc=lambda x: x.vaddr)

    def remove(self, region: R) -> None:
        """
        Remove an existing Region instance from the list.

        :param region: The region to remove.
        """
        if self._is_region_mapped(region):
            self._sorted_list.remove(region)
        self._list.remove(region)

    def find_region_containing(self, addr: int) -> R | None:
        """
        Find the region that contains a specific address. Returns None if none of the regions covers the address.

        :param addr:    The address.
        :return:        The region that covers the specific address, or None if no such region is found.
        """

        pos = key_bisect_find(
            self._sorted_list, addr, keyfunc=lambda x: x if isinstance(x, int) else x.vaddr + x.memsize
        )
        if pos >= len(self._sorted_list):
            return None
        region = self._sorted_list[pos]
        if region.contains_addr(addr):
            return region
        return None

    def find_region_next_to(self, addr: int) -> R | None:
        """
        Find the next region after the given address.

        :param addr: The address to test.
        :return:         The next region that goes after the given address, or None if there is no section after the
                         address,
        """

        pos = key_bisect_find(
            self._sorted_list, addr, keyfunc=lambda x: x if isinstance(x, int) else x.vaddr + x.memsize
        )
        if pos >= len(self._sorted_list):
            return None

        return self._sorted_list[pos]

    @staticmethod
    def _is_region_mapped(region: R) -> bool:
        # delayed import
        # pylint: disable=import-outside-toplevel
        from .elf.regions import ELFSection

        mapped = True
        if region.memsize == 0:
            mapped = False
        elif isinstance(region, ELFSection) and not region.occupies_memory:
            mapped = False
        return mapped

    @staticmethod
    def _make_sorted(lst: list[R]) -> list[R]:
        """
        Return a sorted list of regions that are mapped into memory.

        :param lst:       A list of regions.
        :return:          A sorted list of regions.
        :rtype:           list
        """

        return sorted([r for r in lst if Regions._is_region_mapped(r)], key=lambda x: x.vaddr)



from __future__ import annotations

from cle.address_translator import AT
from cle.backends.backend import register_backend
from cle.backends.elf import ELF
from cle.patched_stream import PatchedStream

ELF_HEADER = bytes.fromhex("7f454c46010101000000000000000000")
CGC_HEADER = bytes.fromhex("7f43474301010143014d6572696e6f00")


class CGC(ELF):
    """
    Backend to support the CGC elf format used by the Cyber Grand Challenge competition.

    See : https://github.com/CyberGrandChallenge/libcgcef/blob/master/cgc_executable_format.md
    """

    is_default = True  # Tell CLE to automatically consider using the CGC backend

    def __init__(self, binary, binary_stream, *args, **kwargs):
        binary_stream = PatchedStream(binary_stream, [(0, ELF_HEADER)])
        super().__init__(binary, binary_stream, *args, **kwargs)
        self.memory.store(AT.from_raw(0, self).to_rva(), CGC_HEADER)  # repair the CGC header
        self.os = "cgc"
        self.execstack = True  # the stack is always executable in CGC

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        identstring = stream.read(4)
        stream.seek(0)
        if identstring.startswith(b"\x7fCGC"):
            return True
        return False

    def _load_segment(self, seg):
        if seg.header.p_memsz > 0:
            super()._load_segment(seg)

    supported_filetypes = ["cgc"]


register_backend("cgc", CGC)



from __future__ import annotations

from .pereloc import PEReloc


class IMAGE_REL_BASED_ARM_MOV32(PEReloc):
    pass


class IMAGE_REL_BASED_THUMB_MOV32(PEReloc):
    pass


relocation_table_arm = {
    5: IMAGE_REL_BASED_ARM_MOV32,
    7: IMAGE_REL_BASED_THUMB_MOV32,
}

__all__ = ("relocation_table_arm",)



from __future__ import annotations

from .backend import ALL_BACKENDS, Backend, ExceptionHandling, FunctionHint, FunctionHintSource, register_backend
from .blob import Blob
from .cartfile import CARTFile
from .cgc import CGC, BackedCGC
from .coff import Coff
from .elf import ELF, ELFCore, MetaELF
from .ihex import Hex
from .java.apk import Apk
from .java.jar import Jar
from .java.soot import Soot
from .macho import MachO
from .minidump import Minidump
from .named_region import NamedRegion
from .pe import PE
from .region import Region, Section, Segment
from .regions import Regions
from .srec import SRec
from .static_archive import StaticArchive
from .symbol import Symbol, SymbolSubType, SymbolType
from .te import TE
from .uefi_firmware import UefiFirmware
from .xbe import XBE

# BinjaBin is not imported by default since importing it is too slow
# you may manually import it by running `from cle.backends.binja import BinjaBin`

__all__ = [
    "FunctionHintSource",
    "FunctionHint",
    "ExceptionHandling",
    "Backend",
    "ALL_BACKENDS",
    "register_backend",
    "ELF",
    "ELFCore",
    "MetaELF",
    "PE",
    "Coff",
    "Blob",
    "CGC",
    "BackedCGC",
    "Hex",
    "SRec",
    "Minidump",
    "MachO",
    "NamedRegion",
    "Jar",
    "Apk",
    "Soot",
    "XBE",
    "StaticArchive",
    "Region",
    "Segment",
    "Section",
    "Regions",
    "Symbol",
    "SymbolType",
    "SymbolSubType",
    "UefiFirmware",
    "TE",
    "CARTFile",
]



# This file is part of Mach-O Loader for CLE.
# Contributed December 2016 by Fraunhofer SIT (https://www.sit.fraunhofer.de/en/).
from __future__ import annotations

import ctypes
import logging
import struct
import sys
import typing
from collections import defaultdict
from io import BufferedReader, BytesIO
from os import SEEK_CUR, SEEK_SET

import archinfo
from sortedcontainers import SortedKeyList

from cle.backends.backend import AT, Backend, register_backend
from cle.backends.macho.binding import BindingHelper, MachOPointerRelocation, MachOSymbolRelocation, read_uleb
from cle.backends.regions import Regions
from cle.errors import CLECompatibilityError, CLEInvalidBinaryError, CLEOperationError

from .encrypted_sentinel_backer import CryptSentinel
from .macho_enums import LoadCommands as LC
from .macho_enums import MachoFiletype, MH_flags
from .section import MachOSection
from .segment import MachOSegment
from .structs import (
    DYLD_CHAINED_PTR_START_NONE,
    ChainedFixupPointerOnDisk,
    DyldChainedPtrFormats,
    DyldImportStruct,
    FileOffset,
    FilePointer,
    MemoryPointer,
    dyld_chained_fixups_header,
    dyld_chained_starts_in_segment,
)
from .symbol import AbstractMachOSymbol, DyldBoundSymbol, SymbolTableSymbol

log = logging.getLogger(name=__name__)

__all__ = ("MachO", "MachOSection", "MachOSegment", "SymbolList")


# pylint: disable=abstract-method
class SymbolList(SortedKeyList):
    """
    Special data structure that extends SortedKeyList to allow looking up a MachO library by name and ordinal quickly
    without having to iterate over the whole list
    """

    _symbol_cache: defaultdict[tuple[str, int], list[AbstractMachOSymbol]]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._symbol_cache = defaultdict(list)

    def add(self, value: AbstractMachOSymbol):
        super().add(value)
        self._symbol_cache[
            (
                value.name,
                value.library_ordinal,
            )
        ].append(value)

    def get_by_name_and_ordinal(self, name: str, ordinal: int, include_stab=False) -> list[AbstractMachOSymbol]:
        if include_stab:
            return self._symbol_cache[(name, ordinal)]
        else:
            return [symbol for symbol in self._symbol_cache[(name, ordinal)] if not symbol.is_stab]


# pylint: enable =abstract-method


class MachO(Backend):
    """
    Mach-O binaries for CLE
    -----------------------

    The Mach-O format is notably different from other formats. Specifically:

    - Sections are always part of a segment, so `self.sections` will be empty.
    - Symbols cannot be categorized like in ELF.
    - Symbol resolution must be handled by the binary.
    - Rebasing in dyld is implemented by adding a small slide to addresses inside the binary, instead of
      changing the base address of the binary. Consequently, the addresses are absolute rather than relative.
      CLE requires relative addresses, leading to numerous `AT.from_lva().to_rva()` calls in this backend.
    """

    is_default = True  # Tell CLE to automatically consider using the MachO backend

    MH_MAGIC_64 = 0xFEEDFACF
    MH_CIGAM_64 = 0xCFFAEDFE
    MH_MAGIC = 0xFEEDFACE
    MH_CIGAM = 0xCEFAEDFE
    ncmds: int
    sizeofcmds: int

    def __init__(self, *args, **kwargs):
        log.warning("The Mach-O backend is not well-supported. Good luck!")

        super().__init__(*args, **kwargs)
        self.symbols = SymbolList(key=self._get_symbol_relative_addr)

        self.struct_byteorder = None  # holds byteorder for struct.unpack(...)
        self._mapped_base = None  # temporary holder für mapped base derived via loading
        self.cputype = None
        self.cpusubtype = None
        self.filetype: int = None
        self.flags = None  # binary flags
        self.imported_libraries = ["Self"]  # ordinal 0 = SELF_LIBRARY_ORDINAL
        self.sections_by_ordinal = [None]  # ordinal 0 = None == Self
        self.exports_by_name = {}  # note exports is currently a raw and unprocessed datastructure.
        # If we intend to use it we must first upgrade it to a class or somesuch
        self.entryoff = None
        self.unixthread_pc = None
        self.os = "macos"
        self.lc_data_in_code = []  # data from LC_DATA_IN_CODE (if encountered). Format: (offset,length,kind)
        self.lc_function_starts: list[int] | None = None  # data from LC_FUNCTION_STARTS (if encountered)
        self.mod_init_func_pointers = []  # may be TUMB interworking
        self.mod_term_func_pointers = []  # may be THUMB interworking
        self.export_blob: bytes | None = None  # exports trie
        self.binding_blob: bytes | None = None  # binding information
        self.lazy_binding_blob: bytes | None = None  # lazy binding information
        self.weak_binding_blob: bytes | None = None  # weak binidng information
        self.rebase_blob: bytes | None = None  # rebasing information
        self.symtab_offset = None  # offset to the symtab
        self.symtab_nsyms = None  # number of symbols in the symtab
        self.binding_done = False  # if true binding was already done and do_bind will be a no-op
        self.strtab: bytes | None = None
        self._indexed_strtab: dict[int, bytes] | None = None
        self._dyld_chained_fixups_offset: int | None = None
        self._dyld_imports: list[AbstractMachOSymbol] = []

        # For some analysis the insertion order of the symbols is relevant and needs to be kept.
        # This is has to be separate from self.symbols because the latter is sorted by address
        self._ordered_symbols: list[AbstractMachOSymbol] = []

        # The minimum version encoded by the LC_BUILD_VERSION command
        self._minimum_version: tuple[int, int, int] | None = None

        # Begin parsing the file
        try:
            binary_file = self._binary_stream
            # get magic value and determine endianness
            self.struct_byteorder = self._detect_byteorder(struct.unpack("=I", binary_file.read(4))[0])

            # parse the mach header:
            # (ignore all irrelevant fields)
            (_, self.cputype, self.cpusubtype, self.filetype, self.ncmds, self.sizeofcmds, self.flags) = self._unpack(
                "7I", binary_file, 0, 28
            )

            # Libraries are always implicitly PIC
            self.pic = bool(self.flags & MH_flags.MH_PIE) or bool(self.filetype & MachoFiletype.MH_DYLIB)

            if not bool(self.flags & MH_flags.MH_TWOLEVEL):  # ensure MH_TWOLEVEL
                log.error(
                    "Binary is not using MH_TWOLEVEL namespacing."
                    "This isn't properly implemented yet and will degrade results in unpredictable ways."
                    "Please open an issue if you encounter this with a binary you can share"
                )

            # determine architecture
            arch_ident = self._detect_arch_ident()
            if not arch_ident:
                raise CLECompatibilityError(f"Unsupported architecture: 0x{self.cputype:X}:0x{self.cpusubtype:X}")

            # Create archinfo
            # Note that this should be customized for Apple ABI (TODO)
            self.set_arch(archinfo.arch_from_id(arch_ident, endness="lsb" if self.struct_byteorder == "<" else "msb"))

            # Determine the base address the binary was linked against
            # and set the values for the Backend and Loader accordingly
            if self.pic and self.filetype == MachoFiletype.MH_EXECUTE:
                assert self.is_main_bin, "An file of type MH_EXECUTE should be the main bin, this should not happen"
                # a Position Independent Main binary would later be loaded at 0x400000, which isn't legal for Mach-O
                # Also, its segment vaddrs are relative to 0x100000000, so we set this as the linked base
                # and the MachO Backend code uses the AdressTranslator to translate linked addresses to relative ones
                # In theory this is the place where the slide for rebasing should be added, but this isn't supported yet
                if self.arch.bits == 64:
                    self.linked_base = self.mapped_base = 2**32
                elif self.arch.bits == 32:
                    self.linked_base = self.mapped_base = 0x4000
            elif self.filetype == MachoFiletype.MH_DYLIB and self.is_main_bin:
                # the segments of dylibs are just relative to the load address, i.e. the lowest segment addr is 0
                # we need to set the load address to something because otherwise the loader will try to map the
                # file to 0x400000, which is technically illegal for Mach-O because of PAGEZERO
                #
                # The problem is that libraries also tend to have relative pointers (e.g. inside ObjC Metadata),
                # which are rebased by parsing the rebase_blob, which isn't supported yet (but coming soon)
                # so we set the base addr to 0 to make them work out without having to deal with this
                # IDA and Ghidra both seem to handle it this way too
                # AFAIU this isn't a problem with iOS15+ binaries anymore that use the new binding fixups
                # but for now we just load all libraries, that are loaded as the main object, at address 0
                #
                # We can't set the linked base to request this, because the MachO Backend implementation
                # uses this to recalculate the addresses
                self._custom_base_addr = 0
            elif self.filetype == MachoFiletype.MH_DYLIB and not self.is_main_bin:
                # A Library is loaded as a dependency, this is fine, the loader will map it to somewhere above the main
                # binary, so we don't need to do anything
                pass
            else:
                # This case is not explicitly supported yet.
                # There are various other MachoFiletypes, which might have different quirks in their loading
                raise CLECompatibilityError(
                    f"Unsupported Mach-O file type: {MachoFiletype(self.filetype)}. "
                    "Please open an issue if you need support for this"
                )

            # Start reading load commands
            lc_offset = (7 if self.arch.bits == 32 else 8) * 4

            self._parse_load_commands(lc_offset)

        except OSError as e:
            log.exception(e)
            raise CLEOperationError(e) from e

        # File is read, begin populating internal fields
        log.info("Parsing exports")
        self._parse_exports()

        if "__mh_execute_header" in self.exports_by_name:
            assert self.exports_by_name["__mh_execute_header"][1] == self.linked_base, (
                "This binary doesn't have a proper __mh_execute_header export, "
                "this breaks assumptions, please report this"
            )

        self._resolve_entry()

        log.info("Parsing %s symbols", self.symtab_nsyms)
        self._parse_symbols(binary_file)
        log.info("Parsing module init/term function pointers")
        self._parse_mod_funcs()

        if self._dyld_chained_fixups_offset:
            log.info("Parsing dyld bound symbols and fixup chains (ios15 and above)")
            self._parse_dyld_chained_fixups()
        else:
            log.info("Parsing binding bytecode stream")
            self.do_binding()

    def set_arch(self, arch):
        super().set_arch(arch)
        self.memory = CryptSentinel(arch=arch)

    @property
    def min_addr(self):
        return self.mapped_base

    @classmethod
    def check_compatibility(cls, spec, obj):
        # TODO: Check properly, but for now libs are just used via force load libs anyway
        return True

    def _parse_load_commands(self, lc_offset):
        # Possible optimization: Remove all unecessary calls to seek()
        # Load commands have a common structure: First 4 bytes identify the command by a magic number
        # second 4 bytes determine the commands size. Everything after this generic "header" is command-specific
        # this makes parsing the commands easy.
        # The documentation for Mach-O is at
        # http://opensource.apple.com//source/xnu/xnu-1228.9.59/EXTERNAL_HEADERS/mach-o/loader.h
        binary_file = self._binary_stream
        count = 0
        offset = lc_offset
        while count < self.ncmds and (offset - lc_offset) < self.sizeofcmds:
            count += 1
            (cmd, size) = self._unpack("II", binary_file, offset, 8)

            # check for segments that interest us
            if cmd in [LC.LC_SEGMENT, LC.LC_SEGMENT_64]:  # LC_SEGMENT,LC_SEGMENT_64
                log.debug("Found LC_SEGMENT(_64) @ %#x", offset)
                self._load_segment(binary_file, offset)
            elif cmd == LC.LC_SYMTAB:  # LC_SYMTAB
                log.debug("Found LC_SYMTAB @ %#x", offset)
                self._load_symtab(binary_file, offset)
            elif cmd in [LC.LC_DYLD_INFO, LC.LC_DYLD_INFO_ONLY]:  # LC_DYLD_INFO(_ONLY)
                log.debug("Found LC_DYLD_INFO(_ONLY) @ %#x", offset)
                self._load_dyld_info(binary_file, offset)
            elif cmd in [LC.LC_LOAD_DYLIB, LC.LC_LOAD_WEAK_DYLIB, LC.LC_REEXPORT_DYLIB]:
                log.debug("Found LC_*_DYLIB @ %#x", offset)
                self._load_dylib_info(binary_file, offset)
            elif cmd == LC.LC_RPATH:  # LC_RPATH
                log.debug("Found LC_RPATH @ %#x", offset)
            elif cmd == LC.LC_MAIN:  # LC_MAIN
                log.debug("Found LC_MAIN @ %#x", offset)
                self._load_lc_main(binary_file, offset)
            elif cmd == LC.LC_UNIXTHREAD:  # LC_UNIXTHREAD
                log.debug("Found LC_UNIXTHREAD @ %#x", offset)
                self._load_lc_unixthread(binary_file, offset)
            elif cmd == LC.LC_FUNCTION_STARTS:  # LC_FUNCTION_STARTS
                log.debug("Found LC_FUNCTION_STARTS @ %#x", offset)
                self._load_lc_function_starts(binary_file, offset)
            elif cmd == LC.LC_DATA_IN_CODE:  # LC_DATA_IN_CODE
                log.debug("Found LC_DATA_IN_CODE @ %#x", offset)
                self._load_lc_data_in_code(binary_file, offset)
            elif cmd in [LC.LC_ENCRYPTION_INFO, LC.LC_ENCRYPTION_INFO_64]:  # LC_ENCRYPTION_INFO(_64)
                log.debug("Found LC_ENCRYPTION_INFO @ %#x", offset)
                # Store the offset and size of the encrypted section for later
                (_, _, cryptoff, cryptsize, cryptid) = self._unpack("5I", binary_file, offset, 20)
                self.memory.set_crypt_info(cryptid, cryptoff, cryptsize)

            elif cmd in [LC.LC_DYLD_CHAINED_FIXUPS]:
                log.info("Found LC_DYLD_CHAINED_FIXUPS @ %#x", offset)
                (_, _, dataoff, datasize) = self._unpack("4I", binary_file, offset, 16)
                self._dyld_chained_fixups_offset: int = dataoff
            elif cmd in [LC.LC_BUILD_VERSION]:
                log.info("Found LC_BUILD_VERSION @ %#x", offset)
                (_, _, _platform, minos, _sdk, _ntools) = self._unpack("6I", binary_file, offset, 6 * 4)
                patch = (minos >> (8 * 0)) & 0xFF
                minor = (minos >> (8 * 1)) & 0xFF
                major = (minos >> (8 * 2)) & 0xFFFF
                self._minimum_version = (major, minor, patch)
                log.info("Found minimum version %s", ".".join([str(i) for i in self._minimum_version]))
            elif cmd in [LC.LC_DYLD_EXPORTS_TRIE]:
                log.info("Found LC_DYLD_EXPORTS_TRIE @ %#x", offset)
                (_, _, dataoff, datasize) = self._unpack("4I", binary_file, offset, 16)
                self.export_blob = self._read(binary_file, dataoff, datasize)
            elif cmd in [LC.LC_DYSYMTAB]:
                # TODO: This probably relevant for library loading and symbols, but it isn't clear how yet
                pass
            else:
                try:
                    command_name = LC(cmd)
                    log.warning("%s is not handled yet", str(command_name))
                except ValueError:
                    log.error("Command %s is not recognized!", hex(cmd))
            # update bookkeeping
            offset += size

        # Assertion to catch malformed binaries - YES this is needed!
        if count < self.ncmds or (offset - lc_offset) < self.sizeofcmds:
            raise CLEInvalidBinaryError(
                f"Assertion triggered: {count} < {self.ncmds} or {offset - lc_offset} < {self.sizeofcmds}"
            )

    @staticmethod
    def extract_soname(path):
        return path.split("/")[-1]

    @classmethod
    def is_compatible(cls, stream):
        stream.seek(0)
        identstring = stream.read(0x5)
        stream.seek(0)
        if (
            identstring.startswith(struct.pack("I", MachO.MH_MAGIC_64))
            or identstring.startswith(struct.pack("I", MachO.MH_CIGAM_64))
            or identstring.startswith(struct.pack("I", MachO.MH_MAGIC))
            or identstring.startswith(struct.pack("I", MachO.MH_CIGAM))
        ):
            return True
        return False

    def is_thumb_interworking(self, address):
        """Returns true if the given address is a THUMB interworking address"""
        # Note: Untested
        return self.arch.bits != 64 and address & 1

    def decode_thumb_interworking(self, address):
        """Decodes a thumb interworking address"""
        # Note: Untested
        return address & ~1 if self.is_thumb_interworking(address) else address

    def _parse_mod_funcs(self):
        log.debug("Parsing module init/term function pointers")

        fmt = "Q" if self.arch.bits == 64 else "I"
        size = 8 if self.arch.bits == 64 else 4

        # factoring out common code
        def parse_mod_funcs_internal(s, target):
            for i in range(s.vaddr, s.vaddr + s.memsize, size):
                rel_address = AT.from_lva(i, self).to_rva()
                addr = self._unpack_with_byteorder(fmt, self.memory.load(rel_address, size))[0]
                log.debug("Addr: %#x", addr)
                target.append(addr)

        for seg in self.segments:
            seg: MachOSection | MachOSegment
            for sec in seg.sections:
                if sec.type == 0x9:  # S_MOD_INIT_FUNC_POINTERS
                    log.debug("Section %s contains init pointers", sec.sectname)
                    parse_mod_funcs_internal(sec, self.mod_init_func_pointers)
                elif sec.type == 0xA:  # S_MOD_TERM_FUNC_POINTERS
                    log.debug("Section %s contains term pointers", sec.sectname)
                    parse_mod_funcs_internal(sec, self.mod_term_func_pointers)

        log.debug("Done parsing module init/term function pointers")

    def find_segment_by_name(self, name):
        for s in self.segments:
            s: MachOSection | MachOSegment
            if s.segname == name:
                return s
        return None

    def _resolve_entry(self):
        if self.entryoff:
            self._entry = self.linked_base + self.entryoff
        elif self.unixthread_pc:
            self._entry = self.unixthread_pc
        else:
            log.warning("No entry point found")
            self._entry = 0

    @staticmethod
    def _read(fp: BufferedReader, offset: int, size: int) -> bytes:
        """
        Simple read abstraction, reads size bytes from offset in file
        :param offset: Offset to seek() to
        :param size: number of bytes to be read
        :return: string of bytes or "" for EOF
        """
        fp.seek(offset)
        return fp.read(size)

    def _unpack_with_byteorder(self, fmt, data) -> tuple[typing.Any, ...]:
        """
        Appends self.struct_byteorder before fmt to ensure usage of correct byteorder
        :return: struct.unpack(self.struct_byteorder+fmt,input)
        """
        return struct.unpack(self.struct_byteorder + fmt, data)

    def _unpack(self, fmt: str, fp: BufferedReader, offset: FilePointer, size: int) -> tuple[typing.Any, ...]:
        """Convenience"""
        return self._unpack_with_byteorder(fmt, self._read(fp, offset, size))

    @staticmethod
    def _detect_byteorder(magic):
        """Determines the binary's byteorder"""

        log.debug("Magic is %#x", magic)

        host_is_little = sys.byteorder == "little"

        if host_is_little:
            if magic in [MachO.MH_MAGIC_64, MachO.MH_MAGIC]:
                log.debug("Detected little-endian")
                return "<"
            elif magic in [MachO.MH_CIGAM, MachO.MH_CIGAM_64]:
                log.debug("Detected big-endian")
                return ">"
            else:
                log.debug("Not a mach-o file")
                raise CLECompatibilityError()
        else:
            if magic in [MachO.MH_MAGIC_64, MachO.MH_MAGIC]:
                log.debug("Detected big-endian")
                return ">"
            elif magic in [MachO.MH_CIGAM_64, MachO.MH_CIGAM]:
                log.debug("Detected little-endian")
                return "<"
            else:
                log.debug("Not a mach-o file")
                raise CLECompatibilityError()

    def do_binding(self):
        # Perform binding

        if self.binding_done:
            log.warning("Binding already done, reset self.binding_done to override if you know what you are doing")
            return

        bh = BindingHelper(self)  # TODO: Make this configurable
        bh.do_normal_bind(self.binding_blob)
        bh.do_lazy_bind(self.lazy_binding_blob)
        bh.do_rebases(self.rebase_blob)
        if self.weak_binding_blob is not None and len(self.weak_binding_blob) > 0:
            log.info(
                "Found weak binding blob. According to current state of knowledge, weak binding "
                "is only sensible if multiple binaries are involved and is thus skipped."
            )

        self.binding_done = True

    def _parse_exports(self):
        """
        Parses the exports trie
        """
        log.debug("Parsing exports")
        blob = self.export_blob
        if blob is None or len(blob) == 0:
            log.debug("Parsing exports done: No exports found")
            return

        # Note some of these fields are currently not used, keep them in to make used variables explicit
        index = 0
        sym_str = b""
        # index,str
        nodes_to_do = [(0, b"")]
        blob_f = BytesIO(blob)  # easier to handle seeking here

        # constants
        # FLAGS_KIND_MASK = 0x03
        # FLAGS_KIND_REGULAR = 0x00
        # FLAGS_KIND_THREAD_LOCAL = 0x01
        # FLAGS_WEAK_DEFINITION = 0x04
        FLAGS_REEXPORT = 0x08
        FLAGS_STUB_AND_RESOLVER = 0x10

        try:
            while True:
                index, sym_str = nodes_to_do.pop()
                log.debug("Processing node %#x %r", index, sym_str)
                blob_f.seek(index, SEEK_SET)
                info_len = struct.unpack("B", blob_f.read(1))[0]
                if info_len > 127:
                    # special case
                    blob_f.seek(-1, SEEK_CUR)
                    tmp = read_uleb(blob, blob_f.tell())  # a bit kludgy
                    info_len = tmp[0]
                    blob_f.seek(tmp[1], SEEK_CUR)

                if info_len > 0:
                    # a symbol is complete
                    tmp = read_uleb(blob, blob_f.tell())
                    blob_f.seek(tmp[1], SEEK_CUR)
                    flags = tmp[0]
                    if flags & FLAGS_REEXPORT:
                        # REEXPORT: uleb:lib ordinal, zero-term str
                        tmp = read_uleb(blob, blob_f.tell())
                        blob_f.seek(tmp[1], SEEK_CUR)
                        lib_ordinal = tmp[0]
                        lib_sym_name = b""
                        char = blob_f.read(1)
                        while char != b"\0":
                            lib_sym_name += char
                            char = blob_f.read(1)
                        log.info("Found REEXPORT export %r: %d,%r", sym_str, lib_ordinal, lib_sym_name)
                        self.exports_by_name[sym_str.decode()] = (flags, lib_ordinal, lib_sym_name.decode())
                    elif flags & FLAGS_STUB_AND_RESOLVER:
                        # STUB_AND_RESOLVER: uleb: stub offset, uleb: resovler offset
                        log.warning("EXPORT: STUB_AND_RESOLVER found")
                        tmp = read_uleb(blob, blob_f.tell())
                        blob_f.seek(tmp[1], SEEK_CUR)
                        stub_offset = tmp[0]
                        tmp = read_uleb(blob, blob_f.tell())
                        blob_f.seek(tmp[1], SEEK_CUR)
                        resolver_offset = tmp[0]
                        log.info("Found STUB_AND_RESOLVER export %r: %#x,%#x'", sym_str, stub_offset, resolver_offset)
                        self.exports_by_name[sym_str.decode()] = (flags, stub_offset, resolver_offset)
                    else:
                        # normal: offset from mach header
                        tmp = read_uleb(blob, blob_f.tell())
                        blob_f.seek(tmp[1], SEEK_CUR)
                        symbol_offset = tmp[0] + self.linked_base
                        log.debug("Found normal export %r: %#x", sym_str, symbol_offset)
                        self.exports_by_name[sym_str.decode()] = (flags, symbol_offset)

                child_count = struct.unpack("B", blob_f.read(1))[0]
                for i in range(0, child_count):
                    child_str = sym_str
                    char = blob_f.read(1)
                    while char != b"\0":
                        child_str += char
                        char = blob_f.read(1)
                    tmp = read_uleb(blob, blob_f.tell())
                    blob_f.seek(tmp[1], SEEK_CUR)
                    next_node = tmp[0]
                    log.debug("%d. child: (%#x, %r)", i, next_node, child_str)
                    nodes_to_do.append((next_node, child_str))

        except IndexError:
            # List is empty we are done!
            log.debug("Done parsing exports")

    def _detect_arch_ident(self):
        """
        Determines the binary's architecture by inspecting cputype and cpusubtype.
        :return: archinfo.arch_from_id-compatible ident string
        """
        # determine architecture by major CPU type
        try:
            arch_lookup = {
                # contains all supported architectures. Note that apple deviates from standard ABI, see Apple docs
                0x100000C: "aarch64",
                0xC: "arm",
                0x7: "x86",
                0x1000007: "x64",
            }
            return arch_lookup[self.cputype]  # subtype currently not needed
        except KeyError:
            return None

    def _load_lc_data_in_code(self, f, off):
        log.debug("Parsing data in code")

        (_, _, dataoff, datasize) = self._unpack("4I", f, off, 16)
        for i in range(dataoff, datasize, 8):
            blob = self._unpack("IHH", f, i, 8)
            self.lc_data_in_code.append(blob)

        log.debug("Done parsing data in code")

    def _load_lc_function_starts(self, f, off):
        # note that the logic below is based on Apple's dyldinfo.cpp, no official docs seem to exist
        log.debug("Parsing function starts")
        (_, _, dataoff, datasize) = self._unpack("4I", f, off, 16)

        i = 0
        end = datasize
        blob = self._read(f, dataoff, datasize)
        self.lc_function_starts = []

        address = None
        for seg in self.segments:
            if seg.offset == 0 and seg.filesize != 0:
                address = seg.vaddr
                break

        if address is None:
            log.error("Could not determine base-address for function starts")
            raise CLEInvalidBinaryError()
        log.debug("Located base-address: %#x", address)

        while i < end:
            uleb = read_uleb(blob, i)

            if blob[i] == 0:
                break  # list is 0 terminated

            address += uleb[0]

            self.lc_function_starts.append(address)
            log.debug("Function start @ %#x (%#x)", uleb[0], address)
            i += uleb[1]
        log.debug("Done parsing function starts")

    def _load_lc_main(self, f, offset):
        if self.entryoff is not None or self.unixthread_pc is not None:
            log.error("More than one entry point for main detected, abort.")
            raise CLEInvalidBinaryError()

        (_, _, self.entryoff, _) = self._unpack("2I2Q", f, offset, 24)
        log.debug("LC_MAIN: entryoff=%#x", self.entryoff)

    def _load_lc_unixthread(self, f, offset):
        if self.entryoff is not None or self.unixthread_pc is not None:
            log.error("More than one entry point for main detected, abort.")
            raise CLEInvalidBinaryError()

        # parse basic structure
        # _, cmdsize, flavor, long_count
        _, _, flavor, _ = self._unpack("4I", f, offset, 16)

        # we only support 4 different types of thread state atm
        # TODO: This is the place to add x86 and x86_64 thread states
        if flavor == 1 and self.arch.bits != 64:  # ARM_THREAD_STATE or ARM_UNIFIED_THREAD_STATE or ARM_THREAD_STATE32
            blob = self._unpack("16I", f, offset + 16, 64)  # parses only until __pc
        elif flavor == 1 and self.arch.bits == 64 or flavor == 6:
            # ARM_THREAD_STATE or ARM_UNIFIED_THREAD_STATE or ARM_THREAD_STATE64
            blob = self._unpack("33Q", f, offset + 16, 264)  # parses only until __pc
        else:
            log.error("Unknown thread flavor: %d", flavor)
            raise CLECompatibilityError()

        self.unixthread_pc = blob[-1]
        log.debug("LC_UNIXTHREAD: __pc=%#x", self.unixthread_pc)

    def _load_dylib_info(self, f, offset):
        (_, _, name_offset, _, _, _) = self._unpack("6I", f, offset, 24)
        lib_path = self.parse_lc_str(f, offset + name_offset)
        log.debug("Adding library %r", lib_path)
        lib_base_name = lib_path.decode("utf-8").rsplit("/", 1)[-1]
        self.deps.append(lib_base_name)
        self.imported_libraries.append(lib_path)

    def _load_dyld_info(self, f: BufferedReader, offset):
        """
        Extracts information blobs for rebasing, binding and export
        """
        (_, _, roff, rsize, boff, bsize, wboff, wbsize, lboff, lbsize, eoff, esize) = self._unpack("12I", f, offset, 48)

        def blob_or_None(f: BufferedReader, off: int, size: int) -> bytes | None:  # helper
            return self._read(f, off, size) if off != 0 and size != 0 else None

        # Extract data blobs
        self.rebase_blob = blob_or_None(f, roff, rsize)
        self.binding_blob = blob_or_None(f, boff, bsize)
        self.weak_binding_blob = blob_or_None(f, wboff, wbsize)
        self.lazy_binding_blob = blob_or_None(f, lboff, lbsize)
        self.export_blob = blob_or_None(f, eoff, esize)

    def _load_symtab(self, f, offset):
        """
        Handles loading of the symbol table
        :param f: input file
        :param offset: offset to the LC_SYMTAB structure
        :return:
        """

        (_, _, symoff, nsyms, stroff, strsize) = self._unpack("6I", f, offset, 24)

        # load string table
        self.strtab = self._read(f, stroff, strsize)

        # Create Dictionary of offsets to strings for quick lookups e.g. during later symbol creation
        _indexed_strtab: dict[int, bytes] = {}
        idx = 0
        for s in self.strtab.split(b"\x00"):
            _indexed_strtab[idx] = s
            idx += len(s) + 1
        self._indexed_strtab = _indexed_strtab

        # store symtab info
        self.symtab_nsyms = nsyms
        self.symtab_offset = symoff

    def _parse_symbols(self, f):
        # parse the symbol entries and create (unresolved) MachOSymbols.
        if self.arch.bits == 64:
            packstr = "I2BHQ"
            structsize = 16
        else:
            packstr = "I2BhI"
            structsize = 12

        for i in range(0, self.symtab_nsyms):
            # The relevant struct is nlist_64 which is defined and documented in mach-o/nlist.h
            offset_in_symtab = i * structsize
            offset = offset_in_symtab + self.symtab_offset
            (n_strx, n_type, n_sect, n_desc, n_value) = self._unpack(packstr, f, offset, structsize)
            log.debug("Adding symbol # %d @ %#x: %s,%s,%s,%s,%s", i, offset, n_strx, n_type, n_sect, n_desc, n_value)
            sym = SymbolTableSymbol(self, offset_in_symtab, n_strx, n_type, n_sect, n_desc, n_value)
            self.symbols.add(sym)
            self._ordered_symbols.append(sym)

            log.debug("Symbol # %d @ %#x is '%s'", i, offset, sym.name)

    def get_string(self, start):
        """Loads a string from the string table"""

        if start in self._indexed_strtab:
            return self._indexed_strtab[start]

        end = start
        if end > len(self.strtab):
            raise ValueError()

        while end < len(self.strtab):
            if self.strtab[end] == 0:
                return self.strtab[start:end]
            end += 1
        return self.strtab[start:]

    def parse_lc_str(self, f, start, limit: int | None = None):
        """Parses a lc_str data structure"""
        tmp = self._unpack("c", f, start, 1)[0]
        s = b""
        ctr = 0
        while tmp != b"\0" and (limit is None or ctr < limit):
            s += tmp
            ctr += 1
            tmp = self._unpack("c", f, start + ctr, 1)[0]

        return s

    def _load_segment(self, f, offset):
        """
        Handles LC_SEGMENT(_64) commands
        :param f: input file
        :param offset: starting offset of the LC_SEGMENT command
        :return:
        """
        # determine if 64 or 32 bit segment
        is64 = self.arch.bits == 64
        if not is64:
            segment_s_size = 56
            (_, _, segname, vmaddr, vmsize, fileoff, filesize, maxprot, initprot, nsects, flags) = self._unpack(
                "2I16s8I", f, offset, segment_s_size
            )
        else:
            segment_s_size = 72
            (_, _, segname, vmaddr, vmsize, fileoff, filesize, maxprot, initprot, nsects, flags) = self._unpack(
                "2I16s4Q4I", f, offset, segment_s_size
            )

        # Cleanup segname
        segname = segname.replace(b"\0", b"")
        log.debug("Processing segment %r", segname)

        # create segment
        seg = MachOSegment(fileoff, vmaddr, filesize, vmsize, segname, nsects, [], flags, initprot, maxprot)

        # Parse section datastructures
        if not is64:
            # 32 bit
            section_s_size = 68
            section_s_packstr = "16s16s9I"
        else:
            # 64 bit
            section_s_size = 80
            # The correct packstring is "16s16s2Q8I", however we use a different one that merges the last two reserved
            # fields (reserved2,reserved3) because it makes the parsing logic below easier
            section_s_packstr = "16s16s2Q6IQ"

        section_start = offset + segment_s_size
        for i in range(0, nsects):
            # Read section
            log.debug("Processing section # %d in %r", i + 1, segname)
            (
                section_sectname,
                section_segname,
                section_vaddr,
                section_vsize,
                section_foff,
                section_align,
                section_reloff,
                section_nreloc,
                section_flags,
                r1,
                r2,
            ) = self._unpack(section_s_packstr, f, (i * section_s_size) + section_start, section_s_size)

            # Clean segname and sectname
            section_sectname = section_sectname.replace(b"\0", b"")
            section_segname = section_segname.replace(b"\0", b"")

            # Create section
            sec = MachOSection(
                section_foff,
                section_vaddr,
                section_vsize,
                section_vsize,
                section_segname,
                section_sectname,
                section_align,
                section_reloff,
                section_nreloc,
                section_flags,
                r1,
                r2,
                parent_segment=seg,
            )

            # Store section
            seg.sections.append(sec)
            self.sections.append(sec)

        # add to sections_by_ordinal
        self.sections_by_ordinal.extend(seg.sections)

        if segname == b"__PAGEZERO":
            # PAGEZERO is a complicated mess so we ignore it entirely
            # It would allocate 4GB of unneeded memory and also break rebasing
            # because now there is a segment that must be at address 0, while the other segments should be slid
            log.info("Found PAGEZERO, skipping backer for memory conservation")
        elif seg.filesize > 0:
            # Append segment data to memory
            blob = self._read(f, seg.offset, seg.filesize)
            if seg.filesize < seg.memsize:
                blob += b"\0" * (seg.memsize - seg.filesize)  # padding

            # The memory of the Backend itself should start at 0, where 0 is the lowest meaningful address
            # In our case this would be the Mach header magic
            # Later this will be loaded at an address like 0x1000000, but that's the job of the loader
            vaddr_offset = AT.from_lva(seg.vaddr, self).to_rva()
            self.memory.add_backer(vaddr_offset, blob)

        # Store segment
        self.segments.append(seg)

    S = typing.TypeVar("S", bound=ctypes.Structure | ctypes.Union)

    def _get_struct(self, struct_type: type[S], offset: int) -> S:
        data = self._read(self._binary_stream, offset, ctypes.sizeof(struct_type))
        return struct_type.from_buffer_copy(data)

    def _read_cstring_from_file(self, start: FilePointer, max_length=None):
        """
        This technically has unnecessary quadratic runtime behavior in `buffer.find` and `buffer+= ...`
        but this shouldn't be noticeable in practice.
        :param start:
        :param max_length:
        :return:
        """
        end = -1
        buffer = b""
        chunk_size = 1024
        while end == -1:
            buffer += self._read(self._binary_stream, start, chunk_size)
            end = buffer.find(b"\x00")
            if end == -1:
                start += chunk_size
            if max_length is not None and len(buffer) > max_length:
                raise ValueError(f"Symbol name exceeds {max_length} bytes, giving up")
        return buffer[:end]

    def _parse_dyld_imports(self, header):
        # Address of Array of dyld_chained_import* structs
        imports_start_addr: FilePointer = self._dyld_chained_fixups_offset + header.imports_offset
        symbols_start_addr: FilePointer = self._dyld_chained_fixups_offset + header.symbols_offset

        import_struct = DyldImportStruct.get_struct(header.imports_format)

        # Parse Imports
        for i in range(header.imports_count):
            import_addr = imports_start_addr + i * ctypes.sizeof(import_struct)
            imp = self._get_struct(import_struct, import_addr)
            sym_name_addr = symbols_start_addr + imp.name_offset
            try:
                sym_name_bytes = self._read_cstring_from_file(sym_name_addr, max_length=2**21)
                sym_name = sym_name_bytes.decode("utf-8")
            except ValueError as e:
                # This symbol string is probably not null terminated, so we can't read it
                log.error("Failed to read symbol name at %x: %s", sym_name_addr, e)
                sym_name = f"<Excessively long symbol name at fileoffset 0x{sym_name_addr:x}>"

            symbols = self.symbols.get_by_name_and_ordinal(sym_name, imp.lib_ordinal)
            if len(symbols) == 1:
                self._dyld_imports.append(symbols[0])
            elif len(symbols) == 0:
                try:
                    log.debug(
                        "Creating DyldBoundSymbol with name %s for library %s",
                        sym_name,
                        self.imported_libraries[imp.lib_ordinal],
                    )
                except IndexError:
                    log.debug(
                        "Creating DyldBoundSymbol with name %s and library ordinal %s (unknown library)",
                        sym_name,
                        imp.lib_ordinal,
                    )
                sym = DyldBoundSymbol(self, sym_name, imp.lib_ordinal)
                self.symbols.add(sym)
                self._dyld_imports.append(sym)
            else:
                raise NotImplementedError(
                    f"Multiple symbols with name {sym_name}" f"for library {self.imported_libraries[imp.lib_ordinal]}."
                )

    def _parse_dyld_chained_fixups(self):
        header: dyld_chained_fixups_header = self._get_struct(
            dyld_chained_fixups_header, self._dyld_chained_fixups_offset
        )

        if header.symbols_format != 0:
            raise NotImplementedError("Dyld fixup symbols are compressed, this isn't supported yet")

        self._parse_dyld_imports(header)

        # Address of the dyld_chained_starts_in_image struct
        segs_addr: FilePointer = self._dyld_chained_fixups_offset + header.starts_offset

        # The struct isn't straightforward to parse with ctypes, so we do it manually
        seg_count = self._unpack("I", self._binary_stream, segs_addr, 4)[0]

        segs: list[FileOffset] = []
        for i in range(seg_count):
            s = self._unpack("I", self._binary_stream, (i * 4) + segs_addr + 4, 4)[0]
            segs.append(s)

            if segs[i] == 0:
                continue

            starts_addr: FilePointer = segs_addr + segs[i]
            starts = self._get_struct(dyld_chained_starts_in_segment, starts_addr)

            seg = self.find_segment_containing(starts.segment_offset)
            # There are weird binaries where the offsets inside the file
            # and inside the virtual addr space don't match anymore.
            # This isn't properly supported yet, and the only known case is the __PII section inside the __ETC segment
            # of rare binaries, which isn't that important for most purposes
            shift = seg.vaddr - (seg.offset)
            if shift != 0:
                assert isinstance(seg, MachOSegment)
                assert seg.segname == "__ETC", (
                    "Only __ETC segments are known to have this shift, please open an"
                    " issue for this binary so it can be investigated"
                )
                log.error("Segment shift detected in, not handling fixups here for now")
                continue

            page_starts_data = self._read(self._binary_stream, starts_addr + 22, starts.page_count * 2)
            page_starts = struct.unpack("<" + ("H" * starts.page_count), page_starts_data)

            pointer_format: DyldChainedPtrFormats = starts.pointer_format
            log.info("Page has pointer_format: %s", pointer_format)
            for j, start in enumerate(page_starts):
                if start == DYLD_CHAINED_PTR_START_NONE:
                    continue
                chain_entry_addr = starts.segment_offset + (j * starts.page_size) + start
                current_chain_addr = chain_entry_addr
                log.info("Reading chain at %x", current_chain_addr)

                while True:
                    chained_rebase_ptr: ChainedFixupPointerOnDisk = self._get_struct(
                        ChainedFixupPointerOnDisk, current_chain_addr
                    )
                    bind = chained_rebase_ptr.isBind(pointer_format)
                    rebase = chained_rebase_ptr.isRebase(pointer_format, self.mapped_base)
                    if bind is not None:
                        libOrdinal, _addend = bind
                        import_symbol = self._dyld_imports[libOrdinal]
                        reloc = MachOSymbolRelocation(self, import_symbol, current_chain_addr, None)
                        self.relocs.append(reloc)
                        # Legacy Code uses bind_xrefs, explicitly add this to make this compatible for now
                        import_symbol.bind_xrefs.append(reloc.dest_addr + self.linked_base)
                        log.debug("Binding for %s found at %x", import_symbol, current_chain_addr)
                    elif rebase is not None:
                        target = self.linked_base + rebase
                        location: MemoryPointer = self.linked_base + current_chain_addr
                        anon_reloc = MachOPointerRelocation(owner=self, relative_addr=current_chain_addr, data=rebase)
                        self.relocs.append(anon_reloc)
                        log.debug("Rebase to %x found at %x", target, location)

                    else:
                        raise CLEInvalidBinaryError("FixupPointer was neither bind nor rebase, that shouldn't happen")

                    skip = chained_rebase_ptr.generic64.rebase.next * 4
                    current_chain_addr += skip
                    if skip == 0:
                        break

    def get_symbol_by_address_fuzzy(self, address):
        """
        Locates a symbol by checking the given address against sym.addr, sym.bind_xrefs and
        sym.symbol_stubs
        """
        for sym in self.symbols:
            if address == sym.relative_addr or address in sym.bind_xrefs or address in sym.symbol_stubs:
                return sym
        return None

    def get_symbol(self, name, include_stab=False, fuzzy=False):  # pylint: disable=arguments-differ
        """
        Returns all symbols matching name.

        Note that especially when include_stab=True there may be multiple symbols with the same
        name, therefore this method always returns an array.

        :param name: the name of the symbol
        :param include_stab: Include debugging symbols NOT RECOMMENDED
        :param fuzzy: Replace exact match with "contains"-style match
        """
        result = []
        for sym in self.symbols:
            if sym.is_stab and not include_stab:
                continue

            if fuzzy:
                if name in sym.name:
                    result.append(sym)
            else:
                if name == sym.name:
                    result.append(sym)

        return result

    def get_symbol_by_insertion_order(self, idx: int) -> AbstractMachOSymbol:
        """

        :param idx: idx when this symbol was inserted
        :return:
        """
        return self._ordered_symbols[idx]

    def get_segment_by_name(self, name):
        """
        Searches for a MachOSegment with the given name and returns it
        :param name: Name of the sought segment
        :return: MachOSegment or None
        """
        for seg in self.segments:
            seg: MachOSection | MachOSegment
            if seg.segname == name:
                return seg

        return None

    def __getitem__(self, item):
        """
        Syntactic sugar for get_segment_by_name
        """
        return self.get_segment_by_name(item)

    segments: Regions[MachOSegment]


register_backend("mach-o", MachO)



from __future__ import annotations

import ntpath
import struct

import archinfo

from cle.backends.backend import Backend, register_backend
from cle.backends.region import Section, Segment
from cle.errors import CLEError, CLEInvalidBinaryError

try:
    from minidump import minidumpfile
    from minidump.streams import SystemInfoStream
except ImportError:
    minidumpfile = None
    SystemInfoStream = None


class MinidumpMissingStreamError(Exception):
    def __init__(self, stream, message=None):
        super().__init__()
        self.message = message
        self.stream = stream


class Minidump(Backend):
    is_default = True

    def __init__(self, *args, **kwargs):
        if minidumpfile is None:
            raise CLEError("Run `pip install minidump==0.0.10` to support loading minidump files")
        super().__init__(*args, **kwargs)
        self.os = "windows"
        self.supports_nx = True
        if self.binary is None:
            self._mdf = minidumpfile.MinidumpFile.parse_bytes(self._binary_stream.read())
        else:
            self._mdf = minidumpfile.MinidumpFile.parse(self.binary)

        self.wow64 = False

        if self._arch is None:
            if getattr(self._mdf, "sysinfo", None) is None:
                raise MinidumpMissingStreamError("SystemInfo", "The architecture was not specified")
            arch = self._mdf.sysinfo.ProcessorArchitecture
            if arch == SystemInfoStream.PROCESSOR_ARCHITECTURE.AMD64:
                if any(module.name.endswith("wow64.dll") for module in self._mdf.modules.modules):
                    self.wow64 = True
                    self.set_arch(archinfo.ArchX86())
                else:
                    self.set_arch(archinfo.ArchAMD64())
            elif arch == SystemInfoStream.PROCESSOR_ARCHITECTURE.INTEL:
                self.set_arch(archinfo.ArchX86())
            else:
                # has not been tested with other architectures
                raise CLEError("Loading minidumps is not implemented for this architecture")

        if self._mdf.memory_segments_64 is not None:
            segments = self._mdf.memory_segments_64.memory_segments
        elif self._mdf.memory_segments is not None:
            segments = self._mdf.memory_segments.memory_segments
        else:
            raise MinidumpMissingStreamError("MemoryList", "The memory segments were not defined")

        for segment in segments:
            data = segment.read(segment.start_virtual_address, segment.size, self._mdf.file_handle)
            self.segments.append(
                Segment(segment.start_file_address, segment.start_virtual_address, segment.size, segment.size)
            )
            self.memory.add_backer(segment.start_virtual_address, data)

        for module in self._mdf.modules.modules:
            for segment in segments:
                if segment.start_virtual_address == module.baseaddress:
                    break
            else:
                raise CLEInvalidBinaryError("Missing segment for loaded module: " + module.name)
            section = Section(module.name, segment.start_file_address, module.baseaddress, module.size)
            self.sections.append(section)
            self.sections_map[ntpath.basename(section.name)] = section

        self._thread_data = {}

        for thread in self._mdf.threads.threads:
            tid = thread.ThreadId
            teb = thread.Teb
            self._binary_stream.seek(thread.ThreadContext.Rva)  # pylint: disable=undefined-loop-variable
            data = self._binary_stream.read(thread.ThreadContext.DataSize)  # pylint: disable=undefined-loop-variable
            self._binary_stream.seek(0)
            self._thread_data[tid] = (teb, data)

    def close(self):
        super().close()
        self._mdf.file_handle.close()
        del self._mdf

    @staticmethod
    def is_compatible(stream):
        identstring = stream.read(4)
        stream.seek(0)
        return identstring == b"MDMP"

    @property
    def threads(self):
        return list(self._thread_data)

    def thread_registers(self, thread=None):
        if thread is None:
            thread = self.threads[0]

        teb, data = self._thread_data[thread]

        if self.arch.name == "AMD64" or self.wow64:
            fmt = "QQQQQQIIHHHHHHIQQQQQQQQQQQQQQQQQQQQQQQ"
            fmt_registers = {
                #'fs':     11, 'gs':  12,
                "eflags": 14,
                "rax": 21,
                "rcx": 22,
                "rdx": 23,
                "rbx": 24,
                "rsp": 25,
                "rbp": 26,
                "rsi": 27,
                "rdi": 28,
                "r8": 29,
                "r9": 30,
                "r10": 31,
                "r11": 32,
                "r12": 33,
                "r13": 34,
                "r14": 35,
                "r15": 36,
                "rip": 37,
            }
        elif self.arch.name == "X86":
            fmt = "IIIIIII112xIIIIIIIIIIIIIIII512x"
            fmt_registers = {
                #'gs':     7,  'fs':  8,
                "edi": 11,
                "esi": 12,
                "ebx": 13,
                "edx": 14,
                "ecx": 15,
                "eax": 16,
                "ebp": 17,
                "eip": 18,
                "eflags": 20,
                "esp": 21,
            }
        else:
            raise CLEError("Deserializing minidump registers is not implemented for this architecture")
        members = struct.unpack_from(fmt, data)
        thread_registers = {}
        for register, position in fmt_registers.items():
            thread_registers[register] = members[position]

        if self.arch.name == "AMD64" or self.wow64:
            gs_base = self.memory.unpack_word(teb + 0x30)
            thread_registers["gs_const"] = gs_base
            if self.arch.name == "AMD64":
                NUM_XMM_REGS = 16
                xmms = struct.unpack_from("16s" * NUM_XMM_REGS, data, offset=0x1A0)
                thread_registers |= {
                    f"xmm{i}": int.from_bytes(xmm, "little", signed=False) for i, xmm in enumerate(xmms)
                }
        elif self.arch.name == "X86":
            fs_base = self.memory.unpack_word(teb + 0x18)
            thread_registers["fs"] = fs_base

        if self.wow64:
            register_translation = [
                ("edi", "rdi"),
                ("esi", "rsi"),
                ("ebx", "rbx"),
                ("edx", "rdx"),
                ("ecx", "rcx"),
                ("eax", "rax"),
                ("ebp", "rbp"),
                ("eip", "rip"),
                ("eflags", "eflags"),
                ("esp", "rsp"),
                ("fs", "gs_const"),  # ???
            ]

            thread_registers = {ereg: thread_registers[rreg] & 0xFFFFFFFF for ereg, rreg in register_translation}

        return thread_registers

    def get_thread_registers_by_id(self, thread_id):
        return self.thread_registers(thread_id)


register_backend("minidump", Minidump)



from __future__ import annotations

import logging

from cle.backends.backend import Backend
from cle.backends.relocation import Relocation
from cle.errors import CLEError

log = logging.getLogger(__name__)


class ThreadManager:
    """
    This class tracks what data is thread-local and can generate thread initialization images

    Most of the heavy lifting will be handled in a subclass
    """

    def __init__(self, loader, arch, max_modules=256):
        self.loader = loader
        self.arch = arch
        self.max_modules = max_modules
        self.modules = []
        self.threads = []

    def register_object(self, obj):
        if not obj.tls_used:
            return False
        if len(self.modules) >= self.max_modules:
            raise CLEError("Too many loaded modules for TLS to handle... file this as a bug")
        obj.tls_module_id = len(self.modules)

        self.modules.append(obj)
        return True

    @staticmethod
    def initialization_image(obj) -> bytes | None:
        if obj.tls_data_start < 0:
            log.warning("The provided object has a negative tls_data_start. Skip TLS loading.")
            return None
        if obj.tls_data_size <= 0:
            log.warning("The provided object has an invalid tls_data_size. Skip TLS loading.")
            return None
        return obj.memory.load(obj.tls_data_start, obj.tls_data_size).ljust(obj.tls_block_size, b"\0")

    def new_thread(self, insert=True):
        thread = self._thread_cls(self)
        if insert:
            self.loader._internal_load(thread)
            self.threads.append(thread)
        return thread

    @property
    def _thread_cls(self):
        raise NotImplementedError("This platform doesn't have an implementation of thread-local storage")


class InternalTLSRelocation(Relocation):
    AUTO_HANDLE_NONE = True

    def __init__(self, val, offset, owner):
        super().__init__(owner, None, offset)
        self.val = val

    @property
    def value(self):
        return self.val + self.owner.mapped_base


class TLSObject(Backend):
    def __init__(self, loader, arch):
        super().__init__("cle##tls", None, loader=loader, arch=arch)



from __future__ import annotations

import logging
import os
from collections import OrderedDict
from enum import Enum

import elftools
import pyvex
from elftools.elf.descriptions import describe_ei_osabi
from elftools.elf.dynamic import DynamicSection
from elftools.elf.enums import ENUM_DT_FLAGS

from cle.address_translator import AT
from cle.backends.backend import Backend
from cle.backends.symbol import SymbolType
from cle.utils import stream_or_path

__all__ = ("MetaELF", "Relro", "maybedecode")

log = logging.getLogger(name=__name__)


class Relro(Enum):
    NONE = 0
    PARTIAL = 1
    FULL = 2


def maybedecode(string):
    # so... it turns out that pyelftools is garbage and will transparently give you either strings or bytestrings
    # based on pretty much nothing whatsoever
    return string if isinstance(string, str) else string.decode()


def _get_relro(elf):
    # The tests for partial and full RELRO have been taken from
    # checksec.sh v1.5 (https://www.trapkit.de/tools/checksec/):
    #   - Partial RELRO has a 'GNU_RELRO' segment
    #   - Full RELRO also has a 'BIND_NOW' flag in the dynamic section
    if not any(seg.header.p_type == "PT_GNU_RELRO" for seg in elf.iter_segments()):
        return Relro.NONE
    dyn_sec = elf.get_section_by_name(".dynamic")
    if dyn_sec is None or not isinstance(dyn_sec, DynamicSection):
        return Relro.PARTIAL
    flags = [tag for tag in dyn_sec.iter_tags() if tag.entry.d_tag == "DT_FLAGS"]
    if len(flags) != 1:
        return Relro.PARTIAL
    return (
        Relro.FULL
        if flags[0].entry.d_val & ENUM_DT_FLAGS["DF_BIND_NOW"] == ENUM_DT_FLAGS["DF_BIND_NOW"]
        else Relro.PARTIAL
    )


class MetaELF(Backend):
    """
    A base class that implements functions used by all backends that can load an ELF.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tmp_reader = elftools.elf.elffile.ELFFile(self._binary_stream)
        self.os = describe_ei_osabi(tmp_reader.header.e_ident.EI_OSABI)
        self.elfflags = tmp_reader.header.e_flags
        self.relro = _get_relro(tmp_reader)
        self._plt = {}
        self._ppc64_abiv1_initial_rtoc = None

        self._cached_plt = None
        self._cached_reverse_plt = None

    supported_filetypes = ["elf"]

    def _block(self, addr, skip_stmts=False):
        # for sanity checking. we live in the world of heuristics now.
        thumb = self.arch.name.startswith("ARM") and addr % 2 == 1
        realaddr = addr
        if thumb:
            realaddr -= 1
        dat = self._block_bytes(realaddr, 40)
        return pyvex.IRSB(dat, addr, self.arch, bytes_offset=1 if thumb else 0, opt_level=1, skip_stmts=skip_stmts)

    def _block_bytes(self, addr, size):
        return self.memory.load(AT.from_lva(addr, self).to_rva(), size)

    def _block_references_addr(self, block, addr):
        if addr in [c.value for c in block.all_constants]:
            return True
        if self.arch.name != "X86":
            return False
        # search for tX = GET(ebx) -> Add32(tX, got_addr - addr)
        if ".got.plt" in self.sections_map:
            got_sec = self.sections_map[".got.plt"]
        elif self.relro is Relro.FULL:
            got_sec = self.sections_map[".got"]
        else:
            return False
        tx = None
        for stmt in block.statements:
            if (
                stmt.tag == "Ist_WrTmp"
                and stmt.data.tag == "Iex_Get"
                and stmt.data.offset == self.arch.registers["ebx"][0]
            ):
                tx = stmt.tmp
            if (
                tx is not None
                and stmt.tag == "Ist_WrTmp"
                and stmt.data.tag == "Iex_Binop"
                and stmt.data.op == "Iop_Add32"
            ):
                args = sorted(stmt.data.args, key=str)
                if (
                    args[0].tag == "Iex_Const"
                    and args[0].con.value == addr - got_sec.vaddr
                    and args[1].tag == "Iex_RdTmp"
                    and args[1].tmp == tx
                ):
                    return True

        return False

    def _add_plt_stub(self, name, addr, sanity_check=True):
        # addr is an LVA
        if addr <= 0:
            return False
        target_addr = self.jmprel[name].linked_addr
        try:
            if sanity_check and not self._block_references_addr(self._block(addr), target_addr):
                return False
        except (pyvex.PyVEXError, KeyError):
            return False
        else:
            self._plt[name] = AT.from_lva(addr, self).to_rva()
            return True

    def _load_plt(self):
        # The main problem here is that there's literally no good way to do this.
        # like, I read through the binutils source and they have a hacked up solution for each arch
        # that performs actual comparisons against the machine code in the plt section.
        # it's pretty bad.
        # we sanity-check all our attempts by requiring that the block lifted at the given address
        # references the GOT slot for the symbol.

        plt_secs = []
        if ".plt" in self.sections_map:
            plt_secs = [self.sections_map[".plt"]]
        if ".plt.got" in self.sections_map:
            plt_secs = [self.sections_map[".plt.got"]]
        if ".MIPS.stubs" in self.sections_map:
            plt_secs.append(self.sections_map[".MIPS.stubs"])
        if ".plt.sec" in self.sections_map:
            plt_secs.append(self.sections_map[".plt.sec"])

        self.jmprel = OrderedDict(sorted(self.jmprel.items(), key=lambda x: x[1].linked_addr))
        func_jmprel = OrderedDict(
            (k, v)
            for k, v in self.jmprel.items()
            if v.symbol.type not in (SymbolType.TYPE_OBJECT, SymbolType.TYPE_SECTION, SymbolType.TYPE_OTHER)
        )

        # ATTEMPT 1: some arches will just leave the plt stub addr in the import symbol
        if self.arch.name in ("ARM", "ARMEL", "ARMHF", "ARMCortexM", "AARCH64", "MIPS32", "MIPS64"):
            for name, reloc in func_jmprel.items():
                if not plt_secs or any(plt_sec.contains_addr(reloc.symbol.linked_addr) for plt_sec in plt_secs):
                    self._add_plt_stub(name, reloc.symbol.linked_addr, sanity_check=bool(plt_secs))

        # ATTEMPT 2: on intel chips the data in the got slot pre-relocation points to a lazy-resolver
        # stub immediately after the plt stub
        if self.arch.name in ("X86", "AMD64"):
            for name, reloc in func_jmprel.items():
                try:
                    self._add_plt_stub(name, self.memory.unpack_word(reloc.relative_addr) - 6, sanity_check=True)
                except KeyError:
                    pass

            # do another sanity check
            if len(set(self._plt.values())) != len(self._plt):
                self._plt = {}

        # ATTEMPT 3: one ppc scheme I've seen is that there are 16-byte stubs packed together
        # right before the resolution stubs.
        if self.arch.name in ("PPC32",):
            resolver_stubs = sorted(
                (self.memory.unpack_word(reloc.relative_addr), name) for name, reloc in func_jmprel.items()
            )
            if resolver_stubs:
                stubs_table = resolver_stubs[0][0] - 16 * len(resolver_stubs)
                for i, (_, name) in enumerate(resolver_stubs):
                    self._add_plt_stub(name, stubs_table + i * 16)

        if len(self._plt) == len(func_jmprel):
            # real quick, bail out before shit hits the fan
            return

        # ATTEMPT 4:
        # ok. time to go in on this.
        # try to find a single plt stub, anywhere. if we already have one, use that, otherwise
        # try to scan forward from _start to __libc_start_main to find that one.
        # then, scan forward and backward from that stub to find the rest of them. yikes!

        # keep a timer so we don't get stuck. keep this short and sweet.
        def tick():
            tick.bailout_timer -= 1
            if tick.bailout_timer <= 0:
                raise TimeoutError()

        tick.bailout_timer = 5

        def scan_forward(addr, name, push=False):
            names = [name] if not isinstance(name, list | tuple) else name

            def block_is_good(blk):
                all_constants = {c.value for c in blk.all_constants}
                for name in names:
                    gotslot = func_jmprel[name].linked_addr
                    if gotslot in all_constants:
                        block_is_good.name = name
                        return True
                return False

            block_is_good.name = None

            def is_endbr(addr):
                if self.arch.name not in ("X86", "AMD64"):
                    return False
                return self._block_bytes(addr, 4) in (b"\xf3\x0f\x1e\xfa", b"\xf3\x0f\x1e\xfb")

            instruction_alignment = self.arch.instruction_alignment
            if self.arch.name in ("ARMEL", "ARMHF"):
                # hard code alignment for ARM code
                instruction_alignment = 4

            try:
                while True:
                    tick()
                    bb = self._block(addr, skip_stmts=False)

                    step_forward = False
                    # the block shouldn't touch any cc_* registers
                    if self.arch.name in ("X86", "AMD64", "ARMEL", "ARMHF", "ARMCortexM"):
                        cc_regs = {
                            self.arch.registers["cc_op"][0],
                            self.arch.registers["cc_ndep"][0],
                            self.arch.registers["cc_dep1"][0],
                            self.arch.registers["cc_dep2"][0],
                        }
                        if any(
                            [isinstance(stmt, pyvex.IRStmt.Put) and stmt.offset in cc_regs for stmt in bb.statements]
                        ):
                            step_forward = True
                        elif any(
                            [
                                isinstance(stmt, pyvex.IRStmt.WrTmp)
                                and isinstance(stmt.data, pyvex.IRExpr.Get)
                                and stmt.data.offset in cc_regs
                                for stmt in bb.statements
                            ]
                        ):
                            step_forward = True

                    if step_forward:
                        # only steps one instruction forward
                        addr += instruction_alignment
                        continue

                    if block_is_good(bb):
                        break
                    if bb.jumpkind == "Ijk_NoDecode":
                        addr += instruction_alignment
                    else:
                        addr += bb.size

                # "push" means try to increase the address as far as we can without regard for semantics
                # the alternative is to only try to lop off nop instructions
                # make sure we don't push through endbr
                if push:
                    if block_is_good.name is None:
                        raise ValueError("block_is_good.name cannot be None.")
                    old_name = block_is_good.name
                    block = self._block(addr, skip_stmts=True)
                    if len(block.instruction_addresses) > 1 and not is_endbr(block.instruction_addresses[0]):
                        for instruction in block.instruction_addresses[1:]:
                            candidate_block = self._block(instruction, skip_stmts=False)
                            if block_is_good(candidate_block) and block_is_good.name == old_name:
                                addr = candidate_block.addr
                                if is_endbr(instruction):
                                    break
                            else:
                                break
                    block_is_good.name = old_name
                else:
                    cont = True
                    while cont:
                        cont = False
                        seen_imark = False
                        # we need to access bb.statements
                        if bb.statements is None:
                            # relift without skipping statements
                            bb = self._block(bb.addr, skip_stmts=False)
                        for stmt in bb.statements:
                            if stmt.tag == "Ist_IMark":
                                if seen_imark:
                                    # good????
                                    bb = self._block(stmt.addr, skip_stmts=False)
                                    if block_is_good(bb):
                                        addr = stmt.addr
                                        cont = True
                                    break
                                else:
                                    seen_imark = True
                            elif stmt.tag == "Ist_Put" and stmt.offset == bb.offsIP:
                                continue
                            else:
                                # there's some behavior, not good
                                break
                return self._add_plt_stub(block_is_good.name, addr)
            except (TimeoutError, ValueError, KeyError, pyvex.PyVEXError):
                return False

        if not self._plt and "__libc_start_main" in func_jmprel and self.entry != 0:
            # try to scan forward through control flow to find __libc_start_main!
            try:
                last_jk = None
                addr = self.entry
                bb = self._block(addr, skip_stmts=True)
                target = bb.default_exit_target
                while target is not None:
                    tick()
                    last_jk = bb.jumpkind
                    addr = target
                    bb = self._block(addr, skip_stmts=True)
                    target = bb.default_exit_target

                if last_jk == "Ijk_Call":
                    self._add_plt_stub("__libc_start_main", addr)
            except (TimeoutError, KeyError, pyvex.PyVEXError):
                pass

        # if func_jmprel.keys()[0] not in self._plt:
        if not set(func_jmprel.keys()).intersection(self._plt.keys()):
            # Check if we have a .plt section
            if not plt_secs:
                # WAHP WAHP
                return

        # some binaries have a bunch of CET stubs before the PLTs, and
        # in the worst case we might have to skip over each one of
        # these... so we set the bailout timer accordingly
        def initial_bailout_timer(func_jmprel):
            return len(func_jmprel) + 5

        if plt_secs:
            # LAST TRY: Find the first block to references ANY GOT slot
            tick.bailout_timer = initial_bailout_timer(func_jmprel)
            scan_forward(min(plt_sec.vaddr for plt_sec in plt_secs), list(func_jmprel.keys()), push=True)

        if not self._plt:
            # \(_^^)/
            return

        # if we've gotten this far there is at least one plt slot address known, guaranteed.
        plt_hitlist = [
            (name, AT.from_rva(self._plt[name], self).to_lva() if name in self._plt else None) for name in func_jmprel
        ]

        name, addr = plt_hitlist[0]
        if addr is None and plt_secs:
            # try to resolve the very first entry
            tick.bailout_timer = initial_bailout_timer(func_jmprel)
            guessed_addr = min(plt_sec.vaddr for plt_sec in plt_secs)
            scan_forward(guessed_addr, name, push=True)
            if name in self._plt:
                # resolved :-)
                plt_hitlist[0] = (name, AT.from_rva(self._plt[name], self).to_lva())

        next_addr = None
        for i, (name, addr) in enumerate(plt_hitlist):
            if addr is None:
                if next_addr is None:
                    continue
                tick.bailout_timer = 5
                scan_forward(next_addr, name, push=True)
                if name in self._plt:
                    addr = AT.from_rva(self._plt[name], self).to_lva()

            if addr is not None:
                try:
                    b0 = self._block(addr, skip_stmts=True)
                except KeyError:
                    # the address does not exist; maybe it's not in the current binary?
                    log.warning("Failed to create a block at address %#x; the address does not exist.", addr)
                    continue
                stub_size = b0.size
                if isinstance(b0.next, pyvex.expr.Const) and b0.next.con.value == addr + b0.size:
                    b1 = self._block(addr + b0.size, skip_stmts=True)
                    stub_size += b1.size
                next_addr = addr + stub_size

    @property
    def plt(self):
        """
        Maps names to addresses.
        """
        if self._cached_plt is None:
            self._cached_plt = {k: AT.from_rva(self._plt[k], self).to_mva() for k in self._plt}
        return self._cached_plt

    @property
    def reverse_plt(self):
        """
        Maps addresses to names.
        """
        if self._cached_reverse_plt is None:
            self._cached_reverse_plt = {AT.from_rva(self._plt[k], self).to_mva(): k for k in self._plt}
        return self._cached_reverse_plt

    @property
    def is_ppc64_abiv1(self):
        """
        Returns whether the arch is PowerPC64 ABIv1.

        :return: True if PowerPC64 ABIv1, False otherwise.
        """
        return self.arch.name == "PPC64" and self.elfflags & 3 < 2

    @property
    def is_ppc64_abiv2(self):
        """
        Returns whether the arch is PowerPC64 ABIv2.

        :return: True if PowerPC64 ABIv2, False otherwise.
        """
        return self.arch.name == "PPC64" and self.elfflags & 3 == 2

    @property
    def ppc64_initial_rtoc(self):
        """
        Get initial rtoc value for PowerPC64 architecture.
        """
        if self.is_ppc64_abiv1:
            return self._ppc64_abiv1_initial_rtoc
        elif self.is_ppc64_abiv2:
            return self._ppc64_abiv2_get_initial_rtoc()
        else:
            return None

    def _ppc64_abiv1_entry_fix(self):
        """
        On PowerPC64, the e_flags elf header entry's lowest two bits determine the ABI type. in ABIv1, the entry point
        given in the elf headers is not actually the entry point, but rather the address in memory where there
        exists a pointer to the entry point.

        Utter bollocks, but this function should fix it.
        """
        if self.is_ppc64_abiv1:
            ep_offset = self._entry
            self._entry = self.memory.unpack_word(AT.from_lva(ep_offset, self).to_rva())
            self._ppc64_abiv1_initial_rtoc = self.memory.unpack_word(AT.from_lva(ep_offset + 8, self).to_rva())

    def _ppc64_abiv2_get_initial_rtoc(self):
        """
        Guess initial table of contents value for PPC64 based on .got section.

        According to PPC64 ABIv2 Specification (Section 3.3): "the TOC pointer
        register typically points to the beginning of the .got section +
        0x8000." Guess the initial rtoc value based on that to handle the
        typical case.
        """
        got_section = self.sections_map.get(".got", None)
        if got_section is None:
            log.warning("Failed to guess initial rtoc value due to missing .got")
            return None
        return got_section.vaddr + 0x8000

    @staticmethod
    def extract_soname(path):
        with stream_or_path(path) as f:
            try:
                e = elftools.elf.elffile.ELFFile(f)
                for seg in e.iter_segments():
                    if seg.header.p_type == "PT_NULL":
                        break
                    elif seg.header.p_type == "PT_DYNAMIC":
                        for tag in seg.iter_tags():
                            if tag.entry.d_tag == "DT_SONAME":
                                return maybedecode(tag.soname)
                        if isinstance(path, str):
                            return os.path.basename(path)

            except elftools.common.exceptions.ELFError:
                pass
            return None



from __future__ import annotations

import logging
import os
import tempfile
from zipfile import ZipFile

from cle.backends.backend import register_backend

from .android_lifecycle import callback
from .soot import Soot

try:
    from pyaxmlparser import APK as APKParser

    PYAXMLPARSER_INSTALLED = True
except ImportError:
    PYAXMLPARSER_INSTALLED = False

try:
    from pysoot.sootir.soot_class import SootClass
    from pysoot.sootir.soot_method import SootMethod
except ImportError:
    SootMethod = None
    SootClass = None

log = logging.getLogger(name=__name__)

# Default list of JNI archs (in descending order of preference)
# => specifies which arch should be used for loading native libs from the APK
default_jni_archs = ["x86", "armeabi", "armeabi-v7a", "x86_64", "arm64-v8a"]


class Apk(Soot):
    """
    Backend for lifting Apk's to Soot.
    """

    is_default = True  # let CLE automatically use this backend

    def __init__(
        self,
        apk_path,
        binary_stream,
        entry_point=None,
        entry_point_params=(),
        android_sdk=None,
        supported_jni_archs=None,
        jni_libs=None,
        jni_libs_ld_path=None,
        **options,
    ):
        """
        :param apk_path:                Path to APK.
        :param android_sdk:             Path to Android SDK folder (e.g. "/home/angr/android/platforms")

        The following parameters are optional

        :param entry_point:             Fully qualified name of method that should be used as the entry point.
        :param supported_jni_archs:     List of supported JNI architectures (ABIs) in descending order of preference.
        :param jni_libs:                Name(s) of JNI libs to load (if any). If not specified, we try to extract
                                        JNI libs from the APK.
        :param jni_libs_ld_path:        Path(s) where to find libs defined by param jni_libs.
                                        Note: Directory of the APK is added by default.
        """

        log.info("Loading APK from %s ...", apk_path)

        if not android_sdk:
            raise ValueError(
                "\nPath to Android SDK must be specified explicitly, e.g.\n"
                '    loading_opts = { "android_sdk" : "/home/angr/android/platforms" }\n'
                '    proj = angr.Project("/path/to/apk/target.apk", main_opts=loading_opts)'
            )

        if not supported_jni_archs:
            supported_jni_archs = default_jni_archs

        # if jni libs are not defined by the user, we try to extract them from the APK
        if not jni_libs:
            log.info("No JNI libs provided. Trying to parse them from the APK.")
            jni_libs, jni_libs_ld_path = self._extract_jni_libs(apk_path, supported_jni_archs)
        else:
            log.info("Using user defined JNI lib(s) %s (load path(s) %s)", jni_libs, jni_libs_ld_path)

        apk_parser = APKParser(apk_path) if PYAXMLPARSER_INSTALLED else None

        if not entry_point:
            if apk_parser:
                main_activity = apk_parser.get_main_activity()
                entry_point = main_activity + "." + "onCreate"
                entry_point_params = ("android.os.Bundle",)
            else:
                log.error("Install pyaxmlparser to identify APK entry point.")
                raise ImportError

        # the actual lifting is done by the Soot superclass
        super().__init__(
            apk_path,
            binary_stream,
            input_format="apk",
            android_sdk=android_sdk,
            entry_point=entry_point,
            entry_point_params=entry_point_params,
            jni_libs=jni_libs,
            jni_libs_ld_path=jni_libs_ld_path,
            **options,
        )

        # the lifecycle needs to support of pyaxmlparser
        if apk_parser:
            self.components = {"activity": [], "service": [], "receiver": [], "provider": []}
            self.callbacks = {"activity": [], "service": [], "receiver": [], "provider": []}
            self._set_lifecycle(apk_parser)
        else:
            self.components = None
            self.callbacks = None
            log.warning("Install pyaxmlparser, if you want to identify components with callbacks.")

    def _set_lifecycle(self, apk_parser):
        """
        Set components with callbacks of APK lifecycle.

        :param pyaxmlparser apk_parser: XML Parser of the APK.
        """

        component_getter = {
            "activity": apk_parser.get_activities,
            "service": apk_parser.get_services,
            "receiver": apk_parser.get_receivers,
            "provider": apk_parser.get_providers,
        }

        for key, getter in component_getter.items():
            class_names = getter()
            self.components[key], self.callbacks[key] = self._extract_lifecycle(class_names, key)

    def _extract_lifecycle(self, cls_name: list[str], component_kind: str) -> tuple[list[SootClass], list[SootMethod]]:
        """
        Extract components with callbacks from class names and component kind.
        Use general callback name for each component by component kind

        :param cls_name:        Name of the class.
        :param component_kind:  Kind of the component. (activity, service, receiver, provider)
        :return components:     The list of class objects which are components.
        :return callbacks:      The list of method objects which are callbacks.
        """

        components = []
        callbacks = []

        for cls in cls_name:
            components.append(self.classes[cls])
            callbacks.extend(self.get_callbacks(cls, callback[component_kind]))

        return components, callbacks

    def get_callbacks(self, class_name: str, callback_names: list[str]) -> list[SootMethod]:
        """
        Get callback methods from the name of callback methods.

        :param class_name:      Name of the class.
        :param callback_names:  Name list of the callbacks.
        :return:                The method object which is callback.
        :rtype:                 list[pysoot.sootir.soot_method.SootMethod]
        """

        callback_methods = []

        for callback_name in callback_names:
            split_str = callback_name.split("(")
            method_name = split_str[0]
            param_str = split_str[1].rstrip(")")

            if param_str == "":
                params = ()
            else:
                params = tuple(param.strip() for param in param_str.split(","))

            soot_method = self.get_soot_method(method_name, class_name=class_name, params=params, none_if_missing=True)
            if soot_method is not None:
                callback_methods.append(soot_method)

        return callback_methods

    @staticmethod
    def _extract_jni_libs(apk_path, supported_jni_archs):
        """
        Extract JNI libs from APK.

        If an APK uses native libraries via JNI, the APK usually include the libs compiled for
        various architectures. This method first matches the available archs with the list of
        supported archs and then extracts the JNI libs using one of the matched archs.

        :return: Name of all extracted JNI libs together with the path to the directory used for
                 extracting.
        :rtype: tuple
        """
        with ZipFile(apk_path) as apk:
            # Step 1: get filelist from APK
            # => structure follows this schema:
            #    AndroidManifest.xml
            #    META-INF/MANIFEST.MF
            #    classes.dex
            #    lib/armeabi-v7a/libnative-lib.so
            #    lib/x86/libnative-lib.so
            filelist = apk.namelist()

            # Step 2: parse name of available libs and archs
            #         from lib paths "/lib/<jni_arch>/lib<name>.so"
            lib_filelist = [list(filter(None, f.split("/"))) for f in filelist if f.startswith("lib")]
            jni_libs = {lib_path[2] for lib_path in lib_filelist if len(lib_path) > 2}
            available_jni_archs = {lib_path[1] for lib_path in lib_filelist if len(lib_path) > 2}

            if not jni_libs:
                log.info("No JNI libs found.")
                return None, None
            log.info("Found JNI lib(s): %s", ", ".join(jni_libs))

            # Step 3: get the first supported jni arch that is available in the APK
            jni_archs = [arch for arch in supported_jni_archs if arch in available_jni_archs]
            if not jni_archs:
                raise ValueError(
                    f"Couldn't find a supported JNI arch. Available {available_jni_archs}. "
                    "Supported {supported_jni_archs}."
                )
            jni_arch = jni_archs[0]
            log.info("Libs are available with arch(s): %s. Picking %s.", ", ".join(available_jni_archs), jni_arch)

            # Step 4: extract all used libaries from the APK
            # TODO: implement this w/o the need of actually writing files to disk
            #       see https://github.com/angr/cle/issues/123
            tmp_dir = tempfile.mkdtemp()
            for lib in jni_libs:
                apk_file = f"lib/{jni_arch}/{lib}"
                apk.extract(apk_file, path=tmp_dir)
            jni_libs_ld_path = os.path.join(tmp_dir, "lib", jni_arch)

            log.info("Extracted lib(s) to %s", jni_libs_ld_path)
            return jni_libs, jni_libs_ld_path

    @staticmethod
    def is_compatible(stream):
        # check if stream is an archive
        if not Soot.is_zip_archive(stream):
            return False
        # get filelist
        with ZipFile(stream) as apk:
            filelist = apk.namelist()
        # check for manifest and the .dex bytecode file
        if "AndroidManifest.xml" not in filelist:
            return False
        if "classes.dex" not in filelist:
            return False
        return True


register_backend("apk", Apk)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from cle.address_translator import AT

from .symbol import Symbol, SymbolType

if TYPE_CHECKING:
    from typing import Any

    from .backend import Backend


log = logging.getLogger(name=__name__)


class Relocation:
    """
    A representation of a relocation in a binary file. Smart enough to
    relocate itself.

    :ivar owner:            The binary this relocation was originally found in, as a cle object
    :ivar symbol:           The Symbol object this relocation refers to
    :ivar relative_addr:    The address in owner this relocation would like to write to
    :ivar resolvedby:       If the symbol this relocation refers to is an import symbol and that import has been
                            resolved, this attribute holds the symbol from a different binary that was used to resolve
                            the import.
    :ivar resolved:         Whether the application of this relocation was successful
    """

    def __init__(self, owner: Backend, symbol: Symbol | None, relative_addr: int):
        self.owner = owner
        self.arch = owner.arch
        self.symbol = symbol
        self.relative_addr = relative_addr
        self.resolvedby: Symbol | None = None
        self.resolved: bool = False
        self.resolvewith = None
        if self.symbol is not None and self.symbol.is_import:
            self.owner.imports[self.symbol.name] = self

    AUTO_HANDLE_NONE = False

    def resolve_symbol(
        self, solist: list[Any], thumb=False, extern_object=None, **kwargs
    ):  # pylint: disable=unused-argument
        if self.resolved:
            return

        if (self.symbol is None or self.symbol.type == SymbolType.TYPE_NONE) and self.AUTO_HANDLE_NONE:
            self.resolve(None, extern_object=extern_object)
            return

        if self.symbol.is_static or self.symbol.is_local:
            # A static or local symbol should only be resolved by itself.
            self.resolve(self.symbol, extern_object=extern_object)
            return

        weak_result = None
        for so in solist:
            symbol = so.get_symbol(self.symbol.name)
            if symbol is not None and symbol.is_export:
                if not symbol.is_weak:
                    self.resolve(symbol, extern_object=extern_object)
                    return
                elif weak_result is None:
                    weak_result = symbol
            # TODO: Was this check obsolted by the addition of is_static?
            # I think right now symbol.is_import = !symbol.is_export
            elif symbol is not None and not symbol.is_import and so is self.owner:
                if not symbol.is_weak:
                    self.resolve(symbol, extern_object=extern_object)
                    return
                elif weak_result is None:
                    weak_result = symbol

        if weak_result is not None:
            self.resolve(weak_result, extern_object=extern_object)
            return

        if self.symbol.is_weak:
            return

        new_symbol = extern_object.make_extern(self.symbol.name, sym_type=self.symbol._type, thumb=thumb)
        self.resolve(new_symbol, extern_object=extern_object)

    def resolve(self, obj, extern_object=None):  # pylint: disable=unused-argument
        self.resolvedby = obj
        self.resolved = True
        if self.symbol is not None:
            if obj is not None:
                log.debug(
                    "%s from %s resolved by %s from %s at %#x",
                    self.symbol.name,
                    self.owner.provides,
                    obj.name,
                    obj.owner.provides,
                    obj.rebased_addr,
                )
            self.symbol.resolve(obj)
        else:
            if not self.AUTO_HANDLE_NONE:
                raise TypeError("Programming error: resolving symbol with None but does not have CAN_HANDLE_NONE set")

    @property
    def rebased_addr(self):
        """
        The address in the global memory space this relocation would like to write to
        """
        return AT.from_rva(self.relative_addr, self.owner).to_mva()

    @property
    def linked_addr(self):
        return AT.from_rva(self.relative_addr, self.owner).to_lva()

    @property
    def dest_addr(self):
        return self.relative_addr

    @property
    def value(self) -> int:  # pylint: disable=no-self-use
        log.error("Value property of Relocation must be overridden by subclass!")
        return 0

    def relocate(self):
        """
        Applies this relocation. Will make changes to the memory object of the
        object it came from.

        This implementation is a generic version that can be overridden in subclasses.
        """
        if not self.resolved:
            return False

        self.owner.memory.pack_word(self.dest_addr, self.value)
        return True

    # compatibility layer

    _complained_owner = False

    @property
    def owner_obj(self):
        if not Relocation._complained_owner:
            Relocation._complained_owner = True
            log.critical("Deprecation warning: use relocation.owner instead of relocation.owner_obj")
        return self.owner



from __future__ import annotations

__all__ = (
    "CLEError",
    "CLEUnknownFormatError",
    "CLEFileNotFoundError",
    "CLEInvalidBinaryError",
    "CLEOperationError",
    "CLECompatibilityError",
    "CLEMemoryError",
)


class CLEError(Exception):
    """
    Base class for errors raised by CLE.
    """

    pass


class CLEUnknownFormatError(CLEError):
    """
    Error raised when CLE encounters an unknown executable file format.
    """

    pass


class CLEFileNotFoundError(CLEError):
    """
    Error raised when a file does not exist.
    """

    pass


class CLEInvalidBinaryError(CLEError):
    """
    Error raised when an executable file is invalid or corrupted.
    """

    pass


class CLEOperationError(CLEError):
    """
    Error raised when a problem is encountered in the process of loading an executable.
    """

    pass


class CLECompatibilityError(CLEError):
    """
    Error raised when loading an executable that is not currently supported by CLE.
    """

    pass


class CLEMemoryError(CLEError):
    """
    Error raised when performing memory operations on unmapped addresses
    """

    pass


class CLEInvalidFileFormatError(CLEError):
    """
    Error raised when loading a file with an invalid format.
    """


class CLEInvalidEncryptionError(CLEError):
    """
    Error raised when loading an encrypted file (e.g., CART) with an incorrect encryption key.
    """

    def __init__(self, backend=None, enckey_argname=None):
        self.backend = backend
        self.enckey_argname = enckey_argname



from __future__ import annotations

import logging

from cle.backends.relocation import Relocation

log = logging.getLogger(name=__name__)


class ELFReloc(Relocation):
    def __init__(self, owner, symbol, relative_addr, addend=None):
        super().__init__(owner, symbol, relative_addr)

        if addend is not None:
            self.is_rela = True
            self._addend = addend
        else:
            self.is_rela = False
            self._addend = self.owner.memory.unpack_word(self.relative_addr)

    @property
    def addend(self):
        if self._addend is None:
            self._addend = self.owner.memory.unpack_word(self.relative_addr)
        return self._addend

    @property
    def value(self):  # pylint: disable=no-self-use
        log.error("Value property of Relocation must be overridden by subclass!")
        return 0



# This file is part of Mach-O Loader for CLE.
# Contributed December 2016 by Fraunhofer SIT (https://www.sit.fraunhofer.de/en/).
from __future__ import annotations

from cle.backends.region import Segment


class MachOSegment(Segment):
    """
    Mach-O Segment

        - offset is the offset into the file the region starts
        - vaddr (or just addr) is the virtual address
        - filesize (or just size) is the size of the region in the file
        - memsize (or vsize) is the size of the region when loaded into memory
        - segname is the segment's name without padding
        - nsect is the number of sections contained in this segment
        - sections is an array of MachOSections
        - flags is a bit vector containing per-segment flags
        - initprot and maxprot are initial and maximum permissions respectively
    """

    def __init__(self, offset, vaddr, size, vsize, segname, nsect, sections, flags, initprot, maxprot):
        super().__init__(offset, vaddr, size, vsize)

        self.segname = segname.decode()
        self.nsect = nsect
        self.sections = sections
        self.flags = flags
        self.initprot = initprot
        self.maxprot = maxprot

    def get_section_by_name(self, name):
        """
        Searches for a section by name within this segment
        :param name: Name of the section
        :return: MachOSection or None
        """
        for sec in self.sections:
            if sec.sectname == name:
                return sec
        return None

    def __getitem__(self, item):
        """
        Syntactic sugar for get_section_by_name
        """
        return self.get_section_by_name(item)

    @property
    def is_readable(self):
        return ((self.initprot | self.maxprot) & 0x01) != 0

    @property
    def is_writable(self):
        return ((self.initprot | self.maxprot) & 0x02) != 0

    @property
    def is_executable(self):
        return ((self.initprot | self.maxprot) & 0x04) != 0

    def __repr__(self):
        return "<{} | offset {:#x}, vaddr {:#x}, size {:#x}>".format(
            self.segname if self.segname else "Unnamed", self.offset, self.vaddr, self.memsize
        )



"""
References:
    - http://www.hexblog.com/wp-content/uploads/2012/06/Recon-2012-Skochinsky-Compiler-Internals.pdf
    - https://www.airs.com/blog/archives/460
    - https://www.airs.com/blog/archives/464
"""

from __future__ import annotations

from elftools.common.utils import struct_parse
from elftools.dwarf.enums import DW_EH_encoding_flags
from elftools.dwarf.structs import DWARFStructs, Struct


class ExceptionTableHeader:
    __slots__ = (
        "lp_start",
        "ttype_encoding",
        "ttype_offset",
        "call_site_encoding",
        "call_site_table_len",
    )

    def __init__(self, lp_start, ttype_encoding, ttype_offset, call_site_encoding, call_site_table_len):
        self.lp_start = lp_start  # landing pad start offset
        self.ttype_encoding = ttype_encoding  # encoding of pointers in type table
        self.ttype_offset = ttype_offset  # Type table offset
        self.call_site_encoding = call_site_encoding  # encoding of items in call site table
        self.call_site_table_len = call_site_table_len  # total length of call site table


class CallSiteEntry:
    __slots__ = ("cs_start", "cs_len", "cs_lp", "cs_action")

    def __init__(self, cs_start, cs_len, cs_lp, cs_action):
        self.cs_start = cs_start
        self.cs_len = cs_len
        self.cs_lp = cs_lp
        self.cs_action = cs_action


class LSDAExceptionTable:
    """
    LSDA exception table parser.

    TODO: Much of this class should be eventually moved to pyelftools.
    """

    def __init__(self, stream, bits, little_endian=True):
        self.address = None
        self.base_offset = None
        self.stream = stream

        if bits in (32, 64):
            dwarf_format = bits
        else:
            raise ValueError(f"Unsupported bits value {bits}. Expect either 32 or 64.")

        self.entry_structs = DWARFStructs(
            little_endian=little_endian, dwarf_format=dwarf_format, address_size=bits // 8
        )
        self._formats = self._eh_encoding_to_field(self.entry_structs)

    @staticmethod
    def _eh_encoding_to_field(entry_structs):
        """
        Shamelessly copied from pyelftools since the original method is a bounded method.

        Return a mapping from basic encodings (DW_EH_encoding_flags) the
        corresponding field constructors (for instance
        entry_structs.Dwarf_uint32).
        """
        return {
            DW_EH_encoding_flags["DW_EH_PE_absptr"]: entry_structs.Dwarf_target_addr,
            DW_EH_encoding_flags["DW_EH_PE_uleb128"]: entry_structs.Dwarf_uleb128,
            DW_EH_encoding_flags["DW_EH_PE_udata2"]: entry_structs.Dwarf_uint16,
            DW_EH_encoding_flags["DW_EH_PE_udata4"]: entry_structs.Dwarf_uint32,
            DW_EH_encoding_flags["DW_EH_PE_udata8"]: entry_structs.Dwarf_uint64,
            DW_EH_encoding_flags["DW_EH_PE_sleb128"]: entry_structs.Dwarf_sleb128,
            DW_EH_encoding_flags["DW_EH_PE_sdata2"]: entry_structs.Dwarf_int16,
            DW_EH_encoding_flags["DW_EH_PE_sdata4"]: entry_structs.Dwarf_int32,
            DW_EH_encoding_flags["DW_EH_PE_sdata8"]: entry_structs.Dwarf_int64,
        }

    def parse_lsda(self, address, offset):
        self.address = address
        self.base_offset = offset
        self.stream.seek(offset)
        header = self._parse_lsda_header()

        csrs: list[CallSiteEntry] = []
        start_offset = self.stream.tell()
        while self.stream.tell() - start_offset < header.call_site_table_len:
            csr = self._parse_call_site_entry(header.call_site_encoding)
            if csr is not None:
                csrs.append(csr)

        return csrs

    def _parse_lsda_header(self):
        # lpstart
        lpstart_encoding = self.stream.read(1)[0]
        if lpstart_encoding != DW_EH_encoding_flags["DW_EH_PE_omit"]:
            base_encoding = lpstart_encoding & 0x0F
            modifier = lpstart_encoding & 0xF0

            lpstart = struct_parse(Struct("dummy", self._formats[base_encoding]("LPStart")), self.stream)["LPStart"]

            if modifier == 0:
                pass
            elif modifier == DW_EH_encoding_flags["DW_EH_PE_pcrel"]:
                lpstart += self.address + (self.stream.tell() - self.base_offset)
            else:
                raise NotImplementedError(f"Unsupported modifier {modifier:#x}.")

        else:
            lpstart = None

        # ttype
        ttype_encoding = self.stream.read(1)[0]
        if ttype_encoding != DW_EH_encoding_flags["DW_EH_PE_omit"]:
            ttype_offset = struct_parse(Struct("dummy", self.entry_structs.Dwarf_uleb128("TType")), self.stream)[
                "TType"
            ]
        else:
            ttype_offset = None

        # call site table length
        cstable_encoding = self.stream.read(1)[0]
        cstable_length = struct_parse(Struct("dummy", self.entry_structs.Dwarf_uleb128("CSTable")), self.stream)[
            "CSTable"
        ]

        return ExceptionTableHeader(
            lpstart,
            ttype_encoding,
            ttype_offset,
            cstable_encoding,
            cstable_length,
        )

    def _parse_call_site_entry(self, encoding):
        base_encoding = encoding & 0x0F
        modifier = encoding & 0xF0

        # header
        s = struct_parse(
            Struct(
                "CallSiteEntry",
                self._formats[base_encoding]("cs_start"),
                self._formats[base_encoding]("cs_len"),
                self._formats[base_encoding]("cs_lp"),
                self.entry_structs.Dwarf_uleb128("cs_action"),
            ),
            self.stream,
        )

        cs_start = s["cs_start"]
        cs_len = s["cs_len"]
        cs_lp = s["cs_lp"]
        cs_action = s["cs_action"]

        if modifier == 0:
            pass
        else:
            raise NotImplementedError(f"Unsupported modifier for CallSiteEntry: {modifier:#x}.")

        return CallSiteEntry(cs_start, cs_len, cs_lp, cs_action)



from __future__ import annotations

import struct


class ELFHashTable:
    """
    Functions to do lookup from a HASH section of an ELF file.

    Information: http://docs.oracle.com/cd/E23824_01/html/819-0690/chapter6-48031.html
    """

    def __init__(self, symtab, stream, offset, arch):
        """
        :param symtab:  The symbol table to perform lookups from (as a pyelftools SymbolTableSection).
        :param stream:  A file-like object to read from the ELF's memory.
        :param offset:  The offset in the object where the table starts.
        :param arch:    The ArchInfo object for the ELF file.
        """
        self.symtab = symtab
        fmt = "<" if arch.memory_endness == "Iend_LE" else ">"
        stream.seek(offset)
        self.nbuckets, self.nchains = struct.unpack(fmt + "II", stream.read(8))
        self.buckets = struct.unpack(fmt + "I" * self.nbuckets, stream.read(4 * self.nbuckets))
        self.chains = struct.unpack(fmt + "I" * self.nchains, stream.read(4 * self.nchains))

    def get(self, k):
        """
        Perform a lookup. Returns a pyelftools Symbol object, or None if there is no match.

        :param k:   The string to look up.
        """
        if self.nbuckets == 0:
            return None, None
        hval = self.elf_hash(k) % self.nbuckets
        symndx = self.buckets[hval]
        while symndx != 0:
            sym = self.symtab.get_symbol(symndx)
            if sym.name == k:
                return symndx, sym
            symndx = self.chains[symndx]
        return None, None

    # from http://www.partow.net/programming/hashfunctions/
    @staticmethod
    def elf_hash(key):
        h = 0
        x = 0
        for c in key:
            h = (h << 4) + ord(c)
            x = h & 0xF0000000
            if x != 0:
                h ^= x >> 24
            h &= ~x
        return h


class GNUHashTable:
    """
    Functions to do lookup from a GNU_HASH section of an ELF file.

    Information: https://blogs.oracle.com/ali/entry/gnu_hash_elf_sections
    """

    def __init__(self, symtab, stream, offset, arch):
        """
        :param symtab:       The symbol table to perform lookups from (as a pyelftools SymbolTableSection).
        :param stream:       A file-like object to read from the ELF's memory.
        :param offset:       The offset in the object where the table starts.
        :param arch:         The ArchInfo object for the ELF file.
        """
        self.symtab = symtab
        fmt = "<" if arch.memory_endness == "Iend_LE" else ">"
        self.c = arch.bits
        fmtsz = "I" if self.c == 32 else "Q"

        stream.seek(offset)
        data = stream.read(16)
        self.nbuckets, self.symndx, self.maskwords, self.shift2 = struct.unpack(fmt + "IIII", data)

        self.bloom = struct.unpack(fmt + fmtsz * self.maskwords, stream.read(self.c * self.maskwords // 8))
        self.buckets = struct.unpack(fmt + "I" * self.nbuckets, stream.read(4 * self.nbuckets))
        self.hash_ptr = stream.tell()
        self.stream = stream

    def _matches_bloom(self, H1):
        C = self.c
        H2 = H1 >> self.shift2
        N = (H1 // C) & (self.maskwords - 1)
        BITMASK = (1 << (H1 % C)) | (1 << (H2 % C))
        return (self.bloom[N] & BITMASK) == BITMASK

    def get(self, k):
        """
        Perform a lookup. Returns a pyelftools Symbol object, or None if there is no match.

        :param k:        The string to look up
        """
        h = self.gnu_hash(k)
        if not self._matches_bloom(h):
            return None, None
        n = self.buckets[h % self.nbuckets]
        if n == 0:
            return None, None
        while True:
            sym = self.symtab.get_symbol(n)
            if sym.name == k:
                return n, sym
            self.stream.seek(self.hash_ptr + 4 * (n - self.symndx))
            if struct.unpack("I", self.stream.read(4))[0] & 1 == 1:
                break
            n += 1
        return None, None

    @staticmethod
    def gnu_hash(key):
        h = 5381
        for c in key:
            h = h * 33 + ord(c)
        return h & 0xFFFFFFFF



from __future__ import annotations

from elftools.elf.enums import ENUM_ST_INFO_TYPE

from cle.address_translator import AT
from cle.backends.symbol import Symbol, SymbolType

from .symbol_type import ELFSymbolType


def maybedecode(string):
    return string if isinstance(string, str) else string.decode()


class ELFSymbol(Symbol):
    """
    Represents a symbol for the ELF format.

    :ivar str binding:      The binding of this symbol as an ELF enum string
    :ivar section:          The section associated with this symbol, or None
    :ivar _subtype:         The ELFSymbolType of this symbol
    """

    def __init__(self, owner, symb):
        subtype_num = ENUM_ST_INFO_TYPE.get(symb.entry.st_info.type, symb.entry.st_info.type)
        arch_list = [owner.arch.name, None]
        if "UNIX" in owner.os:
            arch_list.insert(1, "gnu")
        for arch in arch_list:
            try:
                self._subtype = ELFSymbolType((subtype_num, arch))
            except ValueError:
                pass
            else:
                self._type = self._subtype.to_base_type()
                break
        else:
            self._subtype = None
            self._type = SymbolType.TYPE_OTHER

        sec_ndx, value = symb.entry.st_shndx, symb.entry.st_value

        # A relocatable object's symbol's value is relative to its section's addr.
        if owner.is_relocatable and isinstance(sec_ndx, int):
            value += owner.sections[sec_ndx].remap_offset

        super().__init__(
            owner, maybedecode(symb.name), AT.from_lva(value, owner).to_rva(), symb.entry.st_size, self.type
        )

        self.version = None
        self.binding = symb.entry.st_info.bind
        self.is_hidden = symb.entry["st_other"]["visibility"] == "STV_HIDDEN"
        self.section = sec_ndx if not isinstance(sec_ndx, str) else None
        self.is_static = self._type == SymbolType.TYPE_SECTION or sec_ndx == "SHN_ABS"
        self.is_common = sec_ndx == "SHN_COMMON"
        self.is_weak = self.binding == "STB_WEAK"
        self.is_local = self.binding == "STB_LOCAL"

        self.is_import = sec_ndx == "SHN_UNDEF" and self.binding in ("STB_GLOBAL", "STB_WEAK")
        self.is_export = (self.section is not None or self.is_common) and self.binding in ("STB_GLOBAL", "STB_WEAK")

    @property
    def subtype(self) -> ELFSymbolType:
        return self._subtype
